# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import flask
import mock
import os
import ssl
import threading
import time
import unittest

from Crypto import PublicKey
from jwkest import ecc
from jwkest import jwk
from test import token_utils

from google.api import auth
from google.api.auth import suppliers
from google.api.auth import tokens


class IntegrationTest(unittest.TestCase):

  _CURRENT_TIME = int(time.time())
  _PORT = 8080
  _ISSUER = "https://localhost:%d" % _PORT
  _PROVIDER_ID = "localhost"
  _INVALID_X509_PATH = "invalid-x509"
  _JWKS_PATH = "jwks"
  _SERVICE_NAME = "service@name.com"
  _X509_PATH = "x509"

  _JWT_CLAIMS = {
      "aud": ["https://aud1.local.host", "https://aud2.local.host"],
      "exp": _CURRENT_TIME + 60,
      "email": "user@local.host",
      "iss": _ISSUER,
      "sub": "subject-id"
  }

  _ec_jwk = jwk.ECKey(use="sig").load_key(ecc.P256)
  _rsa_key = jwk.RSAKey(use="sig").load_key(PublicKey.RSA.generate(1024))

  _ec_jwk.kid = "ec-key-id"
  _rsa_key.kid = "rsa-key-id"

  _mock_timer = mock.MagicMock()

  _jwks = jwk.KEYS()
  _jwks._keys.append(_ec_jwk)
  _jwks._keys.append(_rsa_key)

  _AUTH_TOKEN = token_utils.generate_auth_token(_JWT_CLAIMS, _jwks._keys,
                                                alg="RS256", kid=_rsa_key.kid)


  @classmethod
  def setUpClass(cls):
    dirname = os.path.dirname(os.path.realpath(__file__))
    cls._cert_file = os.path.join(dirname, "ssl.cert")
    cls._key_file = os.path.join(dirname, "ssl.key")
    os.environ["REQUESTS_CA_BUNDLE"] = cls._cert_file

    rest_server = cls._RestServer()
    rest_server.start()

  def setUp(self):
    self._provider_ids = {}
    self._configs = {}
    self._authenticator = auth.create_authenticator(self._provider_ids,
                                                    self._configs)

    self._auth_info = mock.MagicMock()
    self._auth_info.is_provider_allowed.return_value = True
    self._auth_info.get_allowed_audiences.return_value = [
        "https://aud1.local.host"
    ]

  def test_verify_auth_token_with_jwks(self):
    url = get_url(IntegrationTest._JWKS_PATH)
    self._provider_ids[self._ISSUER] = self._PROVIDER_ID
    self._configs[IntegrationTest._ISSUER] = suppliers.IssuerUriConfig(False,
                                                                       url)
    user_info = self._authenticator.authenticate(IntegrationTest._AUTH_TOKEN,
                                                 self._auth_info,
                                                 IntegrationTest._SERVICE_NAME)
    self._assert_user_info_equals(tokens.UserInfo(IntegrationTest._JWT_CLAIMS),
                                  user_info)

  def test_authenticate_auth_token_with_bad_signature(self):
    new_rsa_key = jwk.RSAKey(use="sig").load_key(PublicKey.RSA.generate(2048))
    kid = IntegrationTest._rsa_key.kid
    new_rsa_key.kid = kid
    new_jwks = jwk.KEYS()
    new_jwks._keys.append(new_rsa_key)
    auth_token = token_utils.generate_auth_token(IntegrationTest._JWT_CLAIMS,
                                                 new_jwks._keys, alg="RS256",
                                                 kid=kid)
    url = get_url(IntegrationTest._JWKS_PATH)
    self._provider_ids[self._ISSUER] = self._PROVIDER_ID
    self._configs[IntegrationTest._ISSUER] = suppliers.IssuerUriConfig(False,
                                                                       url)
    message = "Signature verification failed"
    with self.assertRaisesRegexp(suppliers.UnauthenticatedException, message):
      self._authenticator.authenticate(auth_token, self._auth_info,
                                       IntegrationTest._SERVICE_NAME)

  def test_verify_auth_token_with_x509(self):
    url = get_url(IntegrationTest._X509_PATH)
    self._provider_ids[self._ISSUER] = self._PROVIDER_ID
    self._configs[IntegrationTest._ISSUER] = suppliers.IssuerUriConfig(False,
                                                                       url)
    user_info = self._authenticator.authenticate(IntegrationTest._AUTH_TOKEN,
                                                 self._auth_info,
                                                 IntegrationTest._SERVICE_NAME)
    self._assert_user_info_equals(tokens.UserInfo(IntegrationTest._JWT_CLAIMS),
                                  user_info)

  def test_verify_auth_token_with_invalid_x509(self):
    url = get_url(IntegrationTest._INVALID_X509_PATH)
    self._provider_ids[self._ISSUER] = self._PROVIDER_ID
    self._configs[IntegrationTest._ISSUER] = suppliers.IssuerUriConfig(False,
                                                                       url)
    message = "Cannot load X.509 certificate"
    with self.assertRaisesRegexp(suppliers.UnauthenticatedException, message):
      self._authenticator.authenticate(IntegrationTest._AUTH_TOKEN,
                                       self._auth_info,
                                       IntegrationTest._SERVICE_NAME)

  def test_openid_discovery(self):
    self._provider_ids[self._ISSUER] = self._PROVIDER_ID
    self._configs[IntegrationTest._ISSUER] = suppliers.IssuerUriConfig(True,
                                                                       None)
    user_info = self._authenticator.authenticate(IntegrationTest._AUTH_TOKEN,
                                                 self._auth_info,
                                                 IntegrationTest._SERVICE_NAME)
    self._assert_user_info_equals(tokens.UserInfo(IntegrationTest._JWT_CLAIMS),
                                  user_info)

  def test_openid_discovery_failed(self):
    self._provider_ids[self._ISSUER] = self._PROVIDER_ID
    self._configs[IntegrationTest._ISSUER] = suppliers.IssuerUriConfig(False,
                                                                       None)
    message = ("Cannot find the `jwks_uri` for issuer %s" %
               IntegrationTest._ISSUER)
    with self.assertRaisesRegexp(suppliers.UnauthenticatedException, message):
      self._authenticator.authenticate(IntegrationTest._AUTH_TOKEN,
                                       self._auth_info,
                                       IntegrationTest._SERVICE_NAME)

  def test_authenticate_with_malformed_auth_code(self):
    with self.assertRaisesRegexp(suppliers.UnauthenticatedException,
                                 "Cannot decode the auth token"):
      self._authenticator.authenticate("invalid-auth-code", self._auth_info,
                                       IntegrationTest._SERVICE_NAME)

  def test_authenticate_with_disallowed_issuer(self):
    url = get_url(IntegrationTest._JWKS_PATH)
    self._configs[IntegrationTest._ISSUER] = suppliers.IssuerUriConfig(False,
                                                                       url)
    message = "Unknown issuer: " + self._ISSUER
    with self.assertRaisesRegexp(suppliers.UnauthenticatedException, message):
      self._authenticator.authenticate(IntegrationTest._AUTH_TOKEN,
                                       self._auth_info,
                                       IntegrationTest._SERVICE_NAME)

  def test_authenticate_with_unknown_issuer(self):
    message = ("Cannot find the `jwks_uri` for issuer %s: "
               "either the issuer is unknown") % IntegrationTest._ISSUER
    with self.assertRaisesRegexp(suppliers.UnauthenticatedException, message):
      self._authenticator.authenticate(IntegrationTest._AUTH_TOKEN,
                                       self._auth_info,
                                       IntegrationTest._SERVICE_NAME)

  def test_authenticate_with_invalid_audience(self):
    url = get_url(IntegrationTest._JWKS_PATH)
    self._provider_ids[self._ISSUER] = self._PROVIDER_ID
    self._configs[IntegrationTest._ISSUER] = suppliers.IssuerUriConfig(False,
                                                                       url)
    self._auth_info.get_allowed_audiences.return_value = []
    with self.assertRaisesRegexp(suppliers.UnauthenticatedException,
                                 "Audiences not allowed"):
      self._authenticator.authenticate(IntegrationTest._AUTH_TOKEN,
                                       self._auth_info,
                                       IntegrationTest._SERVICE_NAME)

  @mock.patch("time.time", _mock_timer)
  def test_authenticate_with_expired_auth_token(self):
    url = get_url(IntegrationTest._JWKS_PATH)
    self._provider_ids[self._ISSUER] = self._PROVIDER_ID
    self._configs[IntegrationTest._ISSUER] = suppliers.IssuerUriConfig(False,
                                                                       url)
    IntegrationTest._mock_timer.return_value = 0

    # Create an auth token that expires in 10 seconds.
    jwt_claims = copy.deepcopy(IntegrationTest._JWT_CLAIMS)
    jwt_claims["exp"] = time.time() + 10
    auth_token = token_utils.generate_auth_token(jwt_claims,
                                                 IntegrationTest._jwks._keys,
                                                 alg="RS256",
                                                 kid=IntegrationTest._rsa_key.kid)

    # Verify that the auth token can be authenticated successfully.
    self._authenticator.authenticate(IntegrationTest._AUTH_TOKEN,
                                     self._auth_info,
                                     IntegrationTest._SERVICE_NAME)

    # Advance the timer by 20 seconds and make sure the token is expired.
    IntegrationTest._mock_timer.return_value += 20
    message = "The auth token has already expired"
    with self.assertRaisesRegexp(suppliers.UnauthenticatedException, message):
      self._authenticator.authenticate(auth_token, self._auth_info,
                                       IntegrationTest._SERVICE_NAME)

  def test_invalid_openid_discovery_url(self):
    issuer = "https://invalid.issuer"
    self._provider_ids[self._ISSUER] = self._PROVIDER_ID
    self._configs[issuer] = suppliers.IssuerUriConfig(True, None)

    jwt_claims = copy.deepcopy(IntegrationTest._JWT_CLAIMS)
    jwt_claims["iss"] = issuer
    auth_token = token_utils.generate_auth_token(jwt_claims,
                                                 IntegrationTest._jwks._keys,
                                                 alg="RS256",
                                                 kid=IntegrationTest._rsa_key.kid)
    message = "Cannot discover the jwks uri"
    with self.assertRaisesRegexp(suppliers.UnauthenticatedException, message):
      self._authenticator.authenticate(auth_token, self._auth_info,
                                       IntegrationTest._SERVICE_NAME)

  def test_invalid_jwks_uri(self):
    url = "https://invalid.jwks.uri"
    self._provider_ids[self._ISSUER] = self._PROVIDER_ID
    self._configs[IntegrationTest._ISSUER] = suppliers.IssuerUriConfig(False,
                                                                       url)
    message = "Cannot retrieve valid verification keys from the `jwks_uri`"
    with self.assertRaisesRegexp(suppliers.UnauthenticatedException, message):
      self._authenticator.authenticate(IntegrationTest._AUTH_TOKEN,
                                       self._auth_info,
                                       IntegrationTest._SERVICE_NAME)

  def _assert_user_info_equals(self, expected, actual):
    self.assertEqual(expected.audiences, actual.audiences)
    self.assertEqual(expected.email, actual.email)
    self.assertEqual(expected.subject_id, actual.subject_id)
    self.assertEqual(expected.issuer, actual.issuer)


  class _RestServer(object):

    def __init__(self):
      app = flask.Flask("integration-test-server")

      @app.route("/" + IntegrationTest._JWKS_PATH)
      def get_json_web_key_set():  # pylint: disable=unused-variable
        return IntegrationTest._jwks.dump_jwks()

      @app.route("/" + IntegrationTest._X509_PATH)
      def get_x509_certificates():  # pylint: disable=unused-variable
        cert = IntegrationTest._rsa_key.key.publickey().exportKey("PEM")
        return flask.jsonify({IntegrationTest._rsa_key.kid: cert})

      @app.route("/" + IntegrationTest._INVALID_X509_PATH)
      def get_invalid_x509_certificates():  # pylint: disable=unused-variable
        return flask.jsonify({IntegrationTest._rsa_key.kid: "invalid cert"})

      @app.route("/.well-known/openid-configuration")
      def get_openid_configuration():  # pylint: disable=unused-variable
        return flask.jsonify({"jwks_uri": get_url(IntegrationTest._JWKS_PATH)})

      self._application = app

    def start(self):
      def run_app():
        ssl_context = (IntegrationTest._cert_file, IntegrationTest._key_file)
        self._application.run(port=IntegrationTest._PORT,
                              ssl_context=ssl_context)

      thread = threading.Thread(target=run_app, args=())
      thread.daemon = True
      thread.start()


def get_url(path):
  return "https://localhost:%d/%s" % (IntegrationTest._PORT, path)



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import httmock
import json
import mock
import os
import sys
import unittest

from apitools.base.py import encoding
from google.api.config import service_config
from google.api.control import messages
from oauth2client import client

class ServiceConfigFetchTest(unittest.TestCase):

  _ACCESS_TOKEN = "test_access_token"

  _SERVICE_NAME = "test_service_name"
  _SERVICE_VERSION = "test_service_version"
  _SERVICE_CONFIG_JSON = {
      "name": _SERVICE_NAME,
      "id": _SERVICE_VERSION
  }

  _credentials = mock.MagicMock()
  _get_http_client = mock.MagicMock()

  def setUp(self):
    os.environ["ENDPOINTS_SERVICE_NAME"] = ServiceConfigFetchTest._SERVICE_NAME
    os.environ["ENDPOINTS_SERVICE_VERSION"] = ServiceConfigFetchTest._SERVICE_VERSION

    self._set_up_default_credential()

  def test_no_service_name(self):
    del os.environ["ENDPOINTS_SERVICE_NAME"]
    message = 'The "ENDPOINTS_SERVICE_NAME" environment variable is not set'
    with self.assertRaisesRegexp(ValueError, message):
      service_config.fetch_service_config()

  def test_no_service_version(self):
    del os.environ["ENDPOINTS_SERVICE_VERSION"]
    message = 'The "ENDPOINTS_SERVICE_VERSION" environment variable is not set'
    with self.assertRaisesRegexp(ValueError, message):
      service_config.fetch_service_config()

  @mock.patch("google.api.config.service_config.client.GoogleCredentials",
              _credentials)
  @mock.patch("google.api.config.service_config._get_http_client", _get_http_client)
  def test_fetch_service_config(self):
    mock_response = mock.MagicMock()
    mock_response.status = 200
    mock_response.data = json.dumps(ServiceConfigFetchTest._SERVICE_CONFIG_JSON)
    mock_http_client = mock.MagicMock()
    mock_http_client.request.return_value = mock_response
    ServiceConfigFetchTest._get_http_client.return_value = mock_http_client

    service = encoding.JsonToMessage(messages.Service,
                                     json.dumps(self._SERVICE_CONFIG_JSON))
    self.assertEqual(service, service_config.fetch_service_config())

    template = service_config._SERVICE_MGMT_URL_TEMPLATE
    url = template.format(ServiceConfigFetchTest._SERVICE_NAME,
                          ServiceConfigFetchTest._SERVICE_VERSION)
    headers={"Authorization": "Bearer " + ServiceConfigFetchTest._ACCESS_TOKEN}
    mock_http_client.request.assert_called_once_with("GET", url,
                                                     headers=headers)

  @mock.patch("google.api.config.service_config.client.GoogleCredentials",
              _credentials)
  @mock.patch("google.api.config.service_config._get_http_client", _get_http_client)
  def test_fetch_service_config_failed(self):
    mock_response = mock.MagicMock()
    mock_response.status = 403
    mock_http_client = mock.MagicMock()
    mock_http_client.request.return_value = mock_response
    ServiceConfigFetchTest._get_http_client.return_value = mock_http_client
    with self.assertRaisesRegexp(Exception, "status code 403"):
      service_config.fetch_service_config()

  @mock.patch("google.api.config.service_config.client.GoogleCredentials",
              _credentials)
  @mock.patch("google.api.config.service_config._get_http_client", _get_http_client)
  def test_fetch_service_config_with_wrong_service_name(self):
    mock_response = mock.MagicMock()
    mock_response.status = 200
    config = copy.deepcopy(ServiceConfigFetchTest._SERVICE_CONFIG_JSON)
    config["name"] = "incorrect_service_name"
    mock_response.data = json.dumps(config)
    mock_http_client = mock.MagicMock()
    mock_http_client.request.return_value = mock_response
    ServiceConfigFetchTest._get_http_client.return_value = mock_http_client

    message = ("Unexpected service name in service config: " +
               config["name"])
    with self.assertRaisesRegexp(ValueError, message):
      service_config.fetch_service_config()

  @mock.patch("google.api.config.service_config.client.GoogleCredentials",
              _credentials)
  @mock.patch("google.api.config.service_config._get_http_client", _get_http_client)
  def test_fetch_service_config_with_wrong_service_version(self):
    mock_response = mock.MagicMock()
    mock_response.status = 200
    config = copy.deepcopy(ServiceConfigFetchTest._SERVICE_CONFIG_JSON)
    config["id"] = "incorrect_service_version"
    mock_response.data = json.dumps(config)
    mock_http_client = mock.MagicMock()
    mock_http_client.request.return_value = mock_response
    ServiceConfigFetchTest._get_http_client.return_value = mock_http_client

    message = ("Unexpected service version in service config: " +
               config["id"])
    with self.assertRaisesRegexp(ValueError, message):
      service_config.fetch_service_config()

  def _set_up_default_credential(self):
    default_credential = mock.MagicMock()
    ServiceConfigFetchTest._credentials.get_application_default.return_value \
        = default_credential
    default_credential.create_scoped.return_value = default_credential
    token = ServiceConfigFetchTest._ACCESS_TOKEN
    access_token = client.AccessTokenInfo(access_token=token, expires_in=None)
    default_credential.get_access_token.return_value = access_token



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import datetime

import unittest2
from expects import be_below_or_equal, expect, equal, raise_error

from google.api.control import timestamp


class TestToRfc3339(unittest2.TestCase):
    A_LONG_TIME_AGO = datetime.datetime(1971, 12, 31, 21, 0, 20, 21000)
    TESTS = [
        (A_LONG_TIME_AGO, '1971-12-31T21:00:20.021Z'),
        (A_LONG_TIME_AGO - datetime.datetime(1970, 1, 1),
         '1971-12-31T21:00:20.021Z')
    ]

    def test_should_converts_correctly(self):
        for t in self.TESTS:
            expect(timestamp.to_rfc3339(t[0])).to(equal(t[1]))

    def test_should_fail_on_invalid_input(self):
        testf = lambda: timestamp.to_rfc3339('this will not work')
        expect(testf).to(raise_error(ValueError))


class TestFromRfc3339(unittest2.TestCase):
    TOLERANCE = 10000  # 1e-5 * 1e9
    TESTS = [
        # Simple
        ('1971-12-31T21:00:20.021Z',
         datetime.datetime(1971, 12, 31, 21, 0, 20, 21000)),
        # different timezone
        ('1996-12-19T16:39:57-08:00',
         datetime.datetime(1996, 12, 20, 0, 39, 57, 0)),
        # microseconds
        ('1996-12-19T16:39:57.123456-08:00',
         datetime.datetime(1996, 12, 20, 0, 39, 57, 123456)),
        # Beyond 2038
        ('2100-01-01T00:00:00Z',
         datetime.datetime(2100, 01, 01, 0, 0, 0, 0))
    ]

    NANO_TESTS = [
        # Simple
        ('1971-12-31T21:00:20.021Z',
         (datetime.datetime(1971, 12, 31, 21, 0, 20, 21000), 21000000)),
        # different timezone
        ('1996-12-19T16:39:57-08:00',
         (datetime.datetime(1996, 12, 20, 0, 39, 57, 0), 0)),
        # microseconds
        ('1996-12-19T16:39:57.123456789-08:00',
         (datetime.datetime(1996, 12, 20, 0, 39, 57, 123457), 123456789)),
    ]

    def test_should_convert_correctly_without_nanos(self):
        for t in self.TESTS:
            expect(timestamp.from_rfc3339(t[0])).to(equal(t[1]))

    def test_should_convert_correctly_with_nanos(self):
        for t in self.NANO_TESTS:
            dt, nanos = timestamp.from_rfc3339(t[0], with_nanos=True)
            expect(dt).to(equal(t[1][0]))
            epsilon = abs(nanos - t[1][1])
            # expect(epsilon).to(equal(0))
            expect(epsilon).to(be_below_or_equal(self.TOLERANCE))


class TestCompare(unittest2.TestCase):
    TESTS = [
        # Strings
        ('1971-10-31T21:00:20.021Z', '1971-11-30T21:00:20.021Z', -1),
        ('1971-11-30T21:00:20.021Z', '1971-10-30T21:00:20.021Z', 1),
        ('1971-11-30T21:00:20Z', '1971-11-30T21:00:20Z', 0),
        ('1971-11-30T21:00:20.021Z', '1971-11-30T21:00:20.041Z', -1),
        ('1971-11-30T21:00:20.021Z', '1971-11-30T21:00:20.001Z', 1),
        # Datetimes
        (datetime.datetime(1996, 10, 20, 0, 39, 57, 0),
         datetime.datetime(1996, 11, 20, 0, 39, 57, 0),
         -1),
        (datetime.datetime(1996, 10, 20, 0, 39, 57, 0),
         datetime.datetime(1996, 10, 20, 0, 39, 57, 0),
         0),
        (datetime.datetime(1996, 11, 20, 0, 39, 57, 0),
         datetime.datetime(1996, 10, 20, 0, 39, 57, 0),
         1)
    ]

    def test_should_compare_correctly(self):
        for t in self.TESTS:
            a, b, want = t
            expect(timestamp.compare(a, b)).to(equal(want))

    def test_should_fail_if_inputs_do_not_have_the_same_type(self):
        testf = lambda: timestamp.compare(self.TESTS[0][0],
                                          datetime.datetime.utcnow())
        expect(testf).to(raise_error(ValueError))
        testf = lambda: timestamp.compare(self.TESTS[0],
                                          datetime.datetime.utcnow())
        expect(testf).to(raise_error(ValueError))



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""client provides a complete standalone service control client.

:class:`Client` is a package-level facade that encapsulates all service control
functionality.

The :class:`Loaders` simplify ``Client`` initialization.

``Client`` needs to stop and start a thread to implement its behaviour.  In most
environments, the default thread class is sufficient.  However, on Google App Engine,
it's necessary to use the appengine specific threading class instead.

:func:`use_gae_thread` and `use_default_thread` can be used to change the thread
class used by new instances of `Client`.

Example:

  >>> from google.api.control import client
  >>>
  >>> # use on appengine with package-default settings
  >>> service_name = 'my-appengine-service-name'
  >>> client.use_gae_thread()
  >>> gae_client = client.Loaders.DEFAULT.load(service_name)
  >>> gae_client.start()

"""
from __future__ import absolute_import

from datetime import datetime, timedelta
from enum import Enum
import json
import logging
import os
import threading
import time

from . import api_client, check_request, report_request
from . import USER_AGENT
from google.api.control.caches import CheckOptions, ReportOptions, to_cache_timer
from google.api.control.vendor.py3 import sched


logger = logging.getLogger(__name__)


CONFIG_VAR = 'ENDPOINTS_SERVER_CONFIG_FILE'


def _load_from_well_known_env():
    if CONFIG_VAR not in os.environ:
        logger.info('did not load server config; no environ var %s', CONFIG_VAR)
        return _load_default()
    json_file = os.environ[CONFIG_VAR]
    if not os.path.exists(json_file):
        logger.warn('did not load service; missing config file %s', json_file)
        return _load_default()
    try:
        with open(json_file) as f:
            json_dict = json.load(f)
            check_json = json_dict['checkAggregatorConfig']
            report_json = json_dict['reportAggregatorConfig']
            check_options = CheckOptions(
                num_entries=check_json['cacheEntries'],
                expiration=timedelta(
                    milliseconds=check_json['responseExpirationMs']),
                flush_interval=timedelta(
                    milliseconds=check_json['flushIntervalMs']))
            report_options = ReportOptions(
                num_entries=report_json['cacheEntries'],
                flush_interval=timedelta(
                    milliseconds=report_json['flushIntervalMs']))
            return check_options, report_options
    except (KeyError, ValueError):
        logger.warn('did not load service; bad json config file %s',
                    json_file,
                    exc_info=True)
        return _load_default()


def _load_default():
    return CheckOptions(), ReportOptions()


def _load_no_cache():
    return (CheckOptions(num_entries=-1),
            ReportOptions(num_entries=-1))


class Loaders(Enum):
    """Enumerates the functions used to load clients from server configs."""
    # pylint: disable=too-few-public-methods
    ENVIRONMENT = (_load_from_well_known_env,)
    DEFAULT = (_load_default,)
    NO_CACHE = (_load_no_cache,)

    def __init__(self, load_func):
        """Constructor.

        load_func is used to load a client config
        """
        self._load_func = load_func

    def load(self, service_name, **kw):
        check_opts, report_opts = self._load_func()
        return Client(service_name, check_opts, report_opts, **kw)


_THREAD_CLASS = threading.Thread


def _create_http_transport():
    additional_http_headers = {"user-agent": USER_AGENT}
    do_logging = logger.level <= logging.DEBUG
    return api_client.ServicecontrolV1(
        additional_http_headers=additional_http_headers,
        log_request=do_logging,
        log_response=do_logging)


def _thread_local_http_transport_func():
    local = threading.local()

    def create_transport():
        if not getattr(local, "transport", None):
            local.transport = _create_http_transport()
        return local.transport

    return create_transport


_CREATE_THREAD_LOCAL_TRANSPORT = _thread_local_http_transport_func()


class Client(object):
    """Client is a package-level facade that encapsulates all service control
    functionality.

    Using one of the :class:`Loaders` makes it easy to initialize ``Client``
    instances.

    Example:

      >>> from google.api.control import client
      >>> service_name = 'my-service-name'
      >>>
      >>> # create an scc client using the package default values
      >>> default_client = client.Loaders.DEFAULT.load(service_name)

      >>> # create an scc client by loading configuration from the
      >>> # a JSON file configured by an environment variable
      >>> json_conf_client = client.Loaders.ENVIRONMENT.load(service_name)

    Client is thread-compatible

    """
    # pylint: disable=too-many-instance-attributes, too-many-arguments

    def __init__(self,
                 service_name,
                 check_options,
                 report_options,
                 timer=datetime.utcnow,
                 create_transport=_CREATE_THREAD_LOCAL_TRANSPORT):
        """

        Args:
            service_name (str): the name of the service to be controlled
            check_options (:class:`google.api.control.caches.CheckOptions`):
              configures checking
            report_options (:class:`google.api.control.caches.ReportOptions`):
              configures reporting
            timer (:func[[datetime.datetime]]: used to obtain the current time.
        """
        self._check_aggregator = check_request.Aggregator(service_name,
                                                          check_options,
                                                          timer=timer)
        self._report_aggregator = report_request.Aggregator(service_name,
                                                            report_options,
                                                            timer=timer)
        self._running = False
        self._scheduler = None
        self._stopped = False
        self._timer = timer
        self._thread = None
        self._create_transport = create_transport
        self._lock = threading.RLock()

    def start(self):
        """Starts processing.

        Calling this method

        - starts the thread that regularly flushes all enabled caches.
        - enables the other methods on the instance to be called successfully

        I.e, even when the configuration disables aggregation, it is invalid to
        access the other methods of an instance until ``start`` is called -
        Calls to other public methods will fail with an AssertionError.

        """
        with self._lock:
            if self._running:
                logger.info('%s is already started', self)
                return

            self._stopped = False
            self._running = True
            logger.info('starting thread of type %s to run the scheduler',
                        _THREAD_CLASS)
            self._thread = _THREAD_CLASS(target=self._schedule_flushes)
            try:
                self._thread.start()
            except Exception:  # pylint: disable=broad-except
                logger.warn(
                    'no scheduler thread, scheduler.run() will be invoked by report(...)',
                    exc_info=True)
                self._thread = None
                self._initialize_flushing()

    def stop(self):
        """Halts processing

        This will lead to the reports being flushed, the caches being cleared
        and a stop to the current processing thread.

        """
        with self._lock:
            if self._stopped:
                logger.info('%s is already stopped', self)
                return

            self._flush_all_reports()
            self._stopped = True
            if self._run_scheduler_directly:
                self._cleanup_if_stopped()

            if self._scheduler and self._scheduler.empty():
                # if there are events scheduled, then _running will subsequently
                # be set False by the scheduler thread.  This handles the
                # case where there are no events, e.g because all aggreagation
                # was disabled
                self._running = False
            self._scheduler = None

    def check(self, check_req):
        """Process a check_request.

        The req is first passed to the check_aggregator.  If there is a valid
        cached response, that is returned, otherwise a response is obtained from
        the transport.

        Args:
          check_req (``ServicecontrolServicesCheckRequest``): to be sent to
            the service control service

        Returns:
           ``CheckResponse``: either the cached response if one is applicable
            or a response from making a transport request, or None if
            if the request to the transport fails

        """

        self._assert_is_running()
        res = self._check_aggregator.check(check_req)
        if res:
            logger.debug('using cached check response for %s: %s',
                         check_request, res)
            return res

        # Application code should not fail because check request's don't
        # complete, They should fail open, so here simply log the error and
        # return None to indicate that no response was obtained
        try:
            transport = self._create_transport()
            resp = transport.services.check(check_req)
            self._check_aggregator.add_response(check_req, resp)
            return resp
        except Exception:  # pylint: disable=broad-except
            logger.error('direct send of check request failed %s',
                         check_request, exc_info=True)
            return None

    def report(self, report_req):
        """Processes a report request.

        It will aggregate it with prior report_requests to be send later
        or it will send it immediately if that's appropriate.
        """
        self._assert_is_running()

        # no thread running, run the scheduler to ensure any pending
        # flush tasks are executed.
        if self._run_scheduler_directly:
            self._scheduler.run(blocking=False)

        if not self._report_aggregator.report(report_req):
            logger.info('need to send a report request directly')
            try:
                transport = self._create_transport()
                transport.services.report(report_req)
            except Exception:  # pylint: disable=broad-except
                logger.error('direct send for report request failed',
                             exc_info=True)

    @property
    def _run_scheduler_directly(self):
        return self._running and self._thread is None

    def _assert_is_running(self):
        assert self._running, '%s needs to be running' % (self,)

    def _initialize_flushing(self):
        with self._lock:
            logger.info('created a scheduler to control flushing')
            self._scheduler = sched.scheduler(to_cache_timer(self._timer),
                                              time.sleep)
            logger.info('scheduling initial check and flush')
            self._flush_schedule_check_aggregator()
            self._flush_schedule_report_aggregator()

    def _schedule_flushes(self):
        # the method expects to be run in the thread created in start()
        self._initialize_flushing()
        self._scheduler.run() # should block until self._stopped is set
        logger.info('scheduler.run completed, %s will exit', threading.current_thread())

    def _cleanup_if_stopped(self):
        if not self._stopped:
            return False

        self._check_aggregator.clear()
        self._report_aggregator.clear()
        self._running = False
        return True

    def _flush_schedule_check_aggregator(self):
        if self._cleanup_if_stopped():
            logger.info('did not schedule check flush: client is stopped')
            return

        flush_interval = self._check_aggregator.flush_interval
        if not flush_interval or flush_interval.total_seconds() < 0:
            logger.debug('did not schedule check flush: caching is disabled')
            return

        if self._run_scheduler_directly:
            logger.debug('did not schedule check flush: no scheduler thread')
            return

        logger.debug('flushing the check aggregator')
        transport = self._create_transport()
        for req in self._check_aggregator.flush():
            try:
                resp = transport.services.check(req)
                self._check_aggregator.add_response(req, resp)
            except Exception:  # pylint: disable=broad-except
                logger.error('failed to flush check_req %s', req, exc_info=True)

        # schedule a repeat of this method
        self._scheduler.enter(
            flush_interval.total_seconds(),
            2,  # a higher priority than report flushes
            self._flush_schedule_check_aggregator,
            ()
        )

    def _flush_schedule_report_aggregator(self):
        if self._cleanup_if_stopped():
            logger.info('did not schedule report flush: client is stopped')
            return

        flush_interval = self._report_aggregator.flush_interval
        if not flush_interval or flush_interval.total_seconds() < 0:
            logger.debug('did not schedule report flush: caching is disabled')
            return

        # flush reports and schedule a repeat of this method
        transport = self._create_transport()
        reqs = self._report_aggregator.flush()
        logger.debug("will flush %d report requests", len(reqs))
        for req in reqs:
            try:
                transport.services.report(req)
            except Exception:  # pylint: disable=broad-except
                logger.error('failed to flush report_req %s', req, exc_info=True)

        self._scheduler.enter(
            flush_interval.total_seconds(),
            1,  # a lower priority than check flushes
            self._flush_schedule_report_aggregator,
            ()
        )

    def _flush_all_reports(self):
        all_requests = self._report_aggregator.clear()
        logger.info('flushing all reports (count=%d)', len(all_requests))
        transport = self._create_transport()
        for req in all_requests:
            try:
                transport.services.report(req)
            except Exception:  # pylint: disable=broad-except
                logger.error('failed to flush report_req %s', req, exc_info=True)


def use_default_thread():
    """Makes ``Client``s started after this use the standard Thread class."""
    global _THREAD_CLASS  # pylint: disable=global-statement
    _THREAD_CLASS = threading.Thread


def use_gae_thread():
    """Makes ``Client``s started after this use the appengine thread class."""
    global _THREAD_CLASS  # pylint: disable=global-statement
    try:
        from google.appengine.api.background_thread import background_thread
        _THREAD_CLASS = background_thread.BackgroundThread
    except ImportError:
        logger.error(
            'Could not install appengine background threads!'
            ' Please install the python AppEngine SDK and use this from there'
        )



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import datetime
import httplib
import unittest2
from expects import be_none, equal, expect, raise_error

from apitools.base.py import encoding

from google.api.control import caches, label_descriptor, timestamp
from google.api.control import check_request, messages, metric_value


class TestSign(unittest2.TestCase):

    def setUp(self):
        op = messages.Operation(
            consumerId=_TEST_CONSUMER_ID,
            operationName=_TEST_OP_NAME
        )
        self.test_check_request = messages.CheckRequest(operation=op)
        self.test_op = op

    def test_should_fail_if_operation_is_not_set(self):
        testf = lambda: check_request.sign(messages.CheckRequest())
        expect(testf).to(raise_error(ValueError))

    def test_should_fail_on_invalid_input(self):
        testf = lambda: check_request.sign(None)
        expect(testf).to(raise_error(ValueError))
        testf = lambda: check_request.sign(object())
        expect(testf).to(raise_error(ValueError))

    def test_should_fail_if_operation_has_no_operation_name(self):
        op = messages.Operation(consumerId=_TEST_CONSUMER_ID)
        testf = lambda: check_request.sign(messages.CheckRequest(operation=op))
        expect(testf).to(raise_error(ValueError))

    def test_should_fail_if_operation_has_no_consumer_id(self):
        op = messages.Operation(operationName=_TEST_OP_NAME)
        testf = lambda: check_request.sign(messages.CheckRequest(operation=op))
        expect(testf).to(raise_error(ValueError))

    def test_should_sign_a_valid_check_request(self):
        check_request.sign(self.test_check_request)

    def test_should_change_signature_when_labels_are_added(self):
        without_labels = check_request.sign(self.test_check_request)
        self.test_op.labels = encoding.PyValueToMessage(
            messages.Operation.LabelsValue, {
                'key1': 'value1',
                'key2': 'value2'
            })
        with_labels = check_request.sign(self.test_check_request)
        expect(with_labels).not_to(equal(without_labels))

    def test_should_change_signature_when_metric_values_are_added(self):
        without_mvs = check_request.sign(self.test_check_request)
        self.test_op.metricValueSets = [
            messages.MetricValueSet(
                metricName='a_float',
                metricValues=[
                    metric_value.create(
                        labels={
                            'key1': 'value1',
                            'key2': 'value2'
                        },
                        doubleValue=1.1,
                    ),
                ]
            )
        ]
        with_mvs = check_request.sign(self.test_check_request)
        expect(with_mvs).not_to(equal(without_mvs))

    def test_should_change_signature_quota_properties_are_specified(self):
        without_qprops = check_request.sign(self.test_check_request)
        self.test_op.quotaProperties = messages.QuotaProperties()
        with_qprops = check_request.sign(self.test_check_request)
        expect(with_qprops).not_to(equal(without_qprops))


class TestAggregatorCheck(unittest2.TestCase):
    SERVICE_NAME = 'service.check'
    FAKE_OPERATION_ID = 'service.general.check'

    def setUp(self):
        self.timer = _DateTimeTimer()
        self.agg = check_request.Aggregator(
            self.SERVICE_NAME, caches.CheckOptions())

    def test_should_fail_if_req_is_bad(self):
        testf = lambda: self.agg.check(object())
        expect(testf).to(raise_error(ValueError))
        testf = lambda: self.agg.check(None)
        expect(testf).to(raise_error(ValueError))

    def test_should_fail_if_service_name_does_not_match(self):
        req = _make_test_request(self.SERVICE_NAME + '-will-not-match')
        testf = lambda: self.agg.check(req)
        expect(testf).to(raise_error(ValueError))

    def test_should_fail_if_check_request_is_missing(self):
        req = messages.ServicecontrolServicesCheckRequest(
            serviceName=self.SERVICE_NAME)
        testf = lambda: self.agg.check(req)
        expect(testf).to(raise_error(ValueError))

    def test_should_fail_if_operation_is_missing(self):
        req = messages.ServicecontrolServicesCheckRequest(
            serviceName=self.SERVICE_NAME,
            checkRequest=messages.CheckRequest())
        testf = lambda: self.agg.check(req)
        expect(testf).to(raise_error(ValueError))

    def test_should_return_none_initially_as_req_is_not_cached(self):
        req = _make_test_request(self.SERVICE_NAME)
        fake_response = messages.CheckResponse(
            operationId=self.FAKE_OPERATION_ID)
        agg = self.agg
        expect(agg.check(req)).to(be_none)


class TestAggregatorThatCannotCache(unittest2.TestCase):
    SERVICE_NAME = 'service.no_cache'
    FAKE_OPERATION_ID = 'service.no_cache.op_id'

    def setUp(self):
        # -ve num_entries means no cache is present
        self.agg = check_request.Aggregator(
            self.SERVICE_NAME,
            caches.CheckOptions(num_entries=-1))

    def test_should_not_cache_responses(self):
        req = _make_test_request(self.SERVICE_NAME)
        fake_response = messages.CheckResponse(
            operationId=self.FAKE_OPERATION_ID)
        agg = self.agg
        expect(agg.check(req)).to(be_none)
        agg.add_response(req, fake_response)
        expect(agg.check(req)).to(be_none)
        agg.clear()
        expect(agg.check(req)).to(be_none)

    def test_should_have_empty_flush_response(self):
        expect(len(self.agg.flush())).to(equal(0))

    def test_should_have_none_as_flush_interval(self):
        expect(self.agg.flush_interval).to(be_none)



class _DateTimeTimer(object):
    def __init__(self, auto=False):
        self.auto = auto
        self.time = datetime.datetime.utcfromtimestamp(0)

    def __call__(self):
        if self.auto:
            self.tick()
        return self.time

    def tick(self):
        self.time += datetime.timedelta(seconds=1)


class TestCachingAggregator(unittest2.TestCase):
    SERVICE_NAME = 'service.with_cache'
    FAKE_OPERATION_ID = 'service.with_cache.op_id'

    def setUp(self):
        self.timer = _DateTimeTimer()
        self.expiration = datetime.timedelta(seconds=2)
        options = caches.CheckOptions(
            flush_interval=datetime.timedelta(seconds=1),
            expiration=self.expiration)
        self.agg = check_request.Aggregator(
            self.SERVICE_NAME, options, timer=self.timer)

    def test_should_have_expiration_as_flush_interval(self):
        expect(self.agg.flush_interval).to(equal(self.expiration))

    def test_should_cache_responses(self):
        req = _make_test_request(self.SERVICE_NAME)
        fake_response = messages.CheckResponse(
            operationId=self.FAKE_OPERATION_ID)
        agg = self.agg
        expect(agg.check(req)).to(be_none)
        agg.add_response(req, fake_response)
        expect(agg.check(req)).to(equal(fake_response))

    def test_should_not_cache_requests_with_important_operations(self):
        req = _make_test_request(
            self.SERVICE_NAME,
            importance=messages.Operation.ImportanceValueValuesEnum.HIGH)
        fake_response = messages.CheckResponse(
            operationId=self.FAKE_OPERATION_ID)
        agg = self.agg
        expect(agg.check(req)).to(be_none)
        agg.add_response(req, fake_response)
        expect(agg.check(req)).to(be_none)

    def test_signals_a_resend_on_1st_call_after_flush_interval(self):
        req = _make_test_request(self.SERVICE_NAME)
        fake_response = messages.CheckResponse(
            operationId=self.FAKE_OPERATION_ID)
        agg = self.agg
        expect(agg.check(req)).to(be_none)
        agg.add_response(req, fake_response)
        expect(agg.check(req)).to(equal(fake_response))

        # Now flush interval is reached, but not the response expiry
        self.timer.tick() # now past the flush_interval
        expect(agg.check(req)).to(be_none)  # none signals the resend

        # Until expiry, the response will continue to be returned
        expect(agg.check(req)).to(equal(fake_response))
        expect(agg.check(req)).to(equal(fake_response))

        # Once expired the cached response is no longer returned
        # expire
        self.timer.tick()
        self.timer.tick() # now expired
        expect(agg.check(req)).to(be_none)
        expect(agg.check(req)).to(be_none)  # 2nd check is None as well

    def test_signals_resend_on_1st_call_after_flush_interval_with_errors(self):
        req = _make_test_request(self.SERVICE_NAME)
        failure_code = messages.CheckError.CodeValueValuesEnum.NOT_FOUND
        fake_response = messages.CheckResponse(
            operationId=self.FAKE_OPERATION_ID, checkErrors=[
                messages.CheckError(code=failure_code)
            ])
        agg = self.agg
        expect(agg.check(req)).to(be_none)
        agg.add_response(req, fake_response)
        expect(agg.check(req)).to(equal(fake_response))

        # Now flush interval is reached, but not the response expiry
        self.timer.tick() # now past the flush_interval
        expect(agg.check(req)).to(be_none)  # first response is null

        # until expiry, the response will continue to be returned
        expect(agg.check(req)).to(equal(fake_response))
        expect(agg.check(req)).to(equal(fake_response))

        # expire
        self.timer.tick()
        self.timer.tick() # now expired
        expect(agg.check(req)).to(be_none)
        expect(agg.check(req)).to(be_none) # 2nd check is None as well

    def test_should_extend_expiration_on_receipt_of_a_response(self):
        req = _make_test_request(self.SERVICE_NAME)
        fake_response = messages.CheckResponse(
            operationId=self.FAKE_OPERATION_ID
        )
        agg = self.agg
        expect(agg.check(req)).to(be_none)
        agg.add_response(req, fake_response)
        expect(agg.check(req)).to(equal(fake_response))

        # Now flush interval is reached, but not the response expiry
        self.timer.tick() # now past the flush_interval
        expect(agg.check(req)).to(be_none)  # first response is null

        # until expiry, the response will continue to be returned
        expect(agg.check(req)).to(equal(fake_response))
        expect(agg.check(req)).to(equal(fake_response))

        # add a response as the request expires
        self.timer.tick()
        agg.add_response(req, fake_response)
        # it would have expired, but because the response was added it does not
        expect(agg.check(req)).to(equal(fake_response))
        expect(agg.check(req)).to(equal(fake_response))
        self.timer.tick() # now past the flush interval again
        expect(agg.check(req)).to(be_none)
        expect(agg.check(req)).to(equal(fake_response))

    def test_does_not_flush_request_that_has_not_been_updated(self):
        req = _make_test_request(self.SERVICE_NAME)
        fake_response = messages.CheckResponse(
            operationId=self.FAKE_OPERATION_ID
        )
        agg = self.agg
        expect(agg.check(req)).to(be_none)
        agg.add_response(req, fake_response)
        self.timer.tick() # now past the flush_interval
        expect(len(agg.flush())).to(equal(0)) # nothing expired
        self.timer.tick() # now past expiry
        self.timer.tick() # now past expiry
        expect(agg.check(req)).to(be_none)  # confirm nothing in cache
        expect(agg.check(req)).to(be_none)  # confirm nothing in cache
        expect(len(agg.flush())).to(equal(0)) # no cached check request

    def test_does_flush_requests_that_have_been_updated(self):
        req = _make_test_request(self.SERVICE_NAME)
        fake_response = messages.CheckResponse(
            operationId=self.FAKE_OPERATION_ID
        )
        agg = self.agg
        expect(agg.check(req)).to(be_none)
        agg.add_response(req, fake_response)
        expect(agg.check(req)).to(equal(fake_response))
        self.timer.tick() # now past the flush_interval
        expect(len(agg.flush())).to(equal(0)) # nothing expired
        self.timer.tick() # now past expiry
        self.timer.tick() # now past expiry
        expect(len(agg.flush())).to(equal(1)) # got the cached check request

    def test_should_clear_requests(self):
        req = _make_test_request(self.SERVICE_NAME)
        fake_response = messages.CheckResponse(
            operationId=self.FAKE_OPERATION_ID
        )
        agg = self.agg
        expect(agg.check(req)).to(be_none)
        agg.add_response(req, fake_response)
        expect(agg.check(req)).to(equal(fake_response))
        agg.clear()
        expect(agg.check(req)).to(be_none)
        expect(len(agg.flush())).to(equal(0))


_TEST_CONSUMER_ID = 'testConsumerID'
_TEST_OP_NAME = 'testOperationName'


def _make_test_request(service_name, importance=None):
    if importance is None:
        importance = messages.Operation.ImportanceValueValuesEnum.LOW
    op = messages.Operation(
        consumerId=_TEST_CONSUMER_ID,
        operationName=_TEST_OP_NAME,
        importance=importance
    )
    check_request = messages.CheckRequest(operation=op)
    return messages.ServicecontrolServicesCheckRequest(
        serviceName=service_name,
        checkRequest=check_request)


_WANTED_USER_AGENT = label_descriptor.USER_AGENT
_START_OF_EPOCH = timestamp.to_rfc3339(datetime.datetime(1970, 1, 1, 0, 0, 0))
_TEST_SERVICE_NAME = 'a_service_name'
_INFO_TESTS = [
    (check_request.Info(
        operation_id='an_op_id',
        operation_name='an_op_name',
        referer='a_referer',
        service_name=_TEST_SERVICE_NAME),
     messages.Operation(
         importance=messages.Operation.ImportanceValueValuesEnum.LOW,
         labels = encoding.PyValueToMessage(
             messages.Operation.LabelsValue, {
                 'servicecontrol.googleapis.com/user_agent': _WANTED_USER_AGENT,
                 'servicecontrol.googleapis.com/referer': 'a_referer'
             }),
         operationId='an_op_id',
         operationName='an_op_name',
         startTime=_START_OF_EPOCH,
         endTime=_START_OF_EPOCH)),
    (check_request.Info(
        api_key='an_api_key',
        api_key_valid=True,
        operation_id='an_op_id',
        operation_name='an_op_name',
        referer='a_referer',
        service_name=_TEST_SERVICE_NAME),
     messages.Operation(
         importance=messages.Operation.ImportanceValueValuesEnum.LOW,
         consumerId='api_key:an_api_key',
         labels = encoding.PyValueToMessage(
             messages.Operation.LabelsValue, {
                 'servicecontrol.googleapis.com/user_agent': _WANTED_USER_AGENT,
                 'servicecontrol.googleapis.com/referer': 'a_referer'
             }),
         operationId='an_op_id',
         operationName='an_op_name',
         startTime=_START_OF_EPOCH,
         endTime=_START_OF_EPOCH)),
    (check_request.Info(
        api_key='an_api_key',
        api_key_valid=False,
        client_ip='127.0.0.1',
        consumer_project_id='project_id',
        operation_id='an_op_id',
        operation_name='an_op_name',
        referer='a_referer',
        service_name=_TEST_SERVICE_NAME),
     messages.Operation(
         importance=messages.Operation.ImportanceValueValuesEnum.LOW,
         consumerId='project:project_id',
         labels = encoding.PyValueToMessage(
             messages.Operation.LabelsValue, {
                 'servicecontrol.googleapis.com/caller_ip': '127.0.0.1',
                 'servicecontrol.googleapis.com/user_agent': _WANTED_USER_AGENT,
                 'servicecontrol.googleapis.com/referer': 'a_referer'
             }),
         operationId='an_op_id',
         operationName='an_op_name',
         startTime=_START_OF_EPOCH,
         endTime=_START_OF_EPOCH)),
]
_INCOMPLETE_INFO_TESTS = [
    check_request.Info(
        operation_name='an_op_name',
        service_name=_TEST_SERVICE_NAME),
    check_request.Info(
        operation_id='an_op_id',
        service_name=_TEST_SERVICE_NAME),
    check_request.Info(
        operation_id='an_op_id',
        operation_name='an_op_name')
]


class TestInfo(unittest2.TestCase):

    def test_should_construct_with_no_args(self):
        expect(check_request.Info()).not_to(be_none)

    def test_should_convert_using_as_check_request(self):
        timer = _DateTimeTimer()
        for info, want in _INFO_TESTS:
            got = info.as_check_request(timer=timer)
            expect(got.checkRequest.operation).to(equal(want))
            expect(got.serviceName).to(equal(_TEST_SERVICE_NAME))

    def test_should_fail_as_check_request_on_incomplete_info(self):
        timer = _DateTimeTimer()
        for info in _INCOMPLETE_INFO_TESTS:
            testf = lambda: info.as_check_request(timer=timer)
            expect(testf).to(raise_error(ValueError))


class TestConvertResponse(unittest2.TestCase):
    PROJECT_ID = 'test_convert_response'

    def test_should_be_ok_with_no_errors(self):
        code, message, _ = check_request.convert_response(
            messages.CheckResponse(), self.PROJECT_ID)
        expect(code).to(equal(httplib.OK))
        expect(message).to(equal(''))

    def test_should_include_project_id_in_error_text_when_needed(self):
        resp = messages.CheckResponse(
            checkErrors = [
                messages.CheckError(
                    code=messages.CheckError.CodeValueValuesEnum.PROJECT_DELETED)
            ]
        )
        code, got, _ = check_request.convert_response(resp, self.PROJECT_ID)
        want = 'Project %s has been deleted' % (self.PROJECT_ID,)
        expect(code).to(equal(httplib.FORBIDDEN))
        expect(got).to(equal(want))

    def test_should_include_detail_in_error_text_when_needed(self):
        detail = 'details, details, details'
        resp = messages.CheckResponse(
            checkErrors = [
                messages.CheckError(
                    code=messages.CheckError.CodeValueValuesEnum.IP_ADDRESS_BLOCKED,
                    detail=detail)
            ]
        )
        code, got, _ = check_request.convert_response(resp, self.PROJECT_ID)
        expect(code).to(equal(httplib.FORBIDDEN))
        expect(got).to(equal(detail))


class _DateTimeTimer(object):
    def __init__(self, auto=False):
        self.auto = auto
        self.time = datetime.datetime(1970, 1, 1)

    def __call__(self):
        if self.auto:
            self.tick()
        return self.time

    def tick(self):
        self.time += datetime.timedelta(seconds=1)



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines a dogpile in-memory cache backend that supports size management."""

from dogpile.cache import api
import pylru


class LruBackend(api.CacheBackend):
  """A dogpile.cache backend that uses LRU as the size management."""

  def __init__(self, options):
    """Initializes an LruBackend.

    Args:
      options: a dictionary that contains configuration options.
    """
    capacity = options["capacity"] if "capacity" in options else 200
    self._cache = pylru.lrucache(capacity)

  def get(self, key):
    return self._cache[key] if key in self._cache else api.NO_VALUE

  def set(self, key, value):
    self._cache[key] = value

  def delete(self, key):
    del self._cache[key]



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import sys
import unittest2
from expects import expect, equal, raise_error

from google.api.control import distribution, messages


class TestCreateExponential(unittest2.TestCase):

    def test_should_fail_if_num_finite_buckets_is_bad(self):
        testf = lambda: distribution.create_exponential(0, 1.1, 0.1)
        expect(testf).to(raise_error(ValueError))

    def test_should_fail_if_growth_factor_is_bad(self):
        testf = lambda: distribution.create_exponential(1, 0.9, 0.1)
        expect(testf).to(raise_error(ValueError))

    def test_should_fail_if_scale_is_bad(self):
        testf = lambda: distribution.create_exponential(1, 1.1, -0.1)
        expect(testf).to(raise_error(ValueError))

    def test_should_succeed_if_inputs_are_ok(self):
        num_finite_buckets = 1
        got = distribution.create_exponential(num_finite_buckets, 1.1, 0.1)
        expect(len(got.bucketCounts)).to(equal(num_finite_buckets + 2))


class TestCreateLinear(unittest2.TestCase):

    def test_should_fail_if_num_finite_buckets_is_bad(self):
        testf = lambda: distribution.create_linear(0, 1.1, 0.1)
        expect(testf).to(raise_error(ValueError))

    def test_should_fail_if_growth_factor_is_bad(self):
        testf = lambda: distribution.create_linear(1, -0.1, 0.1)
        expect(testf).to(raise_error(ValueError))

    def test_should_succeed_if_inputs_are_ok(self):
        num_finite_buckets = 1
        got = distribution.create_linear(num_finite_buckets, 0.1, 0.1)
        expect(len(got.bucketCounts)).to(equal(num_finite_buckets + 2))


class TestCreateExplicit(unittest2.TestCase):

    def test_should_fail_if_there_are_matching_bounds(self):
        testf = lambda: distribution.create_explicit([0.0, 0.1, 0.1])
        expect(testf).to(raise_error(ValueError))

    def test_should_succeed_if_inputs_are_ok(self):
        want = [0.1, 0.2, 0.3]
        got = distribution.create_explicit([0.1, 0.2, 0.3])
        expect(got.explicitBuckets.bounds).to(equal(want))
        expect(len(got.bucketCounts)).to(equal(len(want) + 1))

    def test_should_succeed_if_input_bounds_are_unsorted(self):
        want = [0.1, 0.2, 0.3]
        got = distribution.create_explicit([0.3, 0.1, 0.2])
        expect(got.explicitBuckets.bounds).to(equal(want))


def _make_explicit_dist():
    return distribution.create_explicit([0.1, 0.3, 0.5, 0.7])


def _make_linear_dist():
    return distribution.create_linear(3, 0.2, 0.1)


def _make_exponential_dist():
    return distribution.create_exponential(3, 2, 0.1)

_UNDERFLOW_SAMPLE = 1e-5
_LOW_SAMPLE = 0.11
_HIGH_SAMPLE = 0.5
_OVERFLOW_SAMPLE = 1e5

_TEST_SAMPLES_AND_BUCKETS = [
    {
        'samples': [_UNDERFLOW_SAMPLE],
        'want': [1, 0, 0, 0, 0]
    },
    {
        'samples': [_LOW_SAMPLE] * 2,
        'want': [0, 2, 0, 0, 0]
    },
    {
        'samples': [_LOW_SAMPLE, _HIGH_SAMPLE, _HIGH_SAMPLE],
        'want': [0, 1, 0, 2, 0]
    },
    {
        'samples': [_OVERFLOW_SAMPLE],
        'want': [0, 0, 0, 0, 1]
    },
]


def _expect_stats_eq_direct_calc_from_samples(d, samples):
    # pylint: disable=fixme
    # TODO: update this the sum of rho-squared
    want_mean = sum(samples) / len(samples)
    expect(d.mean).to(equal(want_mean))
    expect(d.maximum).to(equal(max(samples)))
    expect(d.minimum).to(equal(min(samples)))


class TestAddSample(unittest2.TestCase):
    NOTHING_SET = messages.Distribution()

    def test_should_fail_if_no_buckets_are_set(self):
        testf = lambda: distribution.add_sample(_UNDERFLOW_SAMPLE,
                                                self.NOTHING_SET)
        expect(testf).to(raise_error(ValueError))

    def expect_adds_test_samples_ok(self, make_dist_func):
        for t in _TEST_SAMPLES_AND_BUCKETS:
            d = make_dist_func()
            samples = t['samples']
            for s in samples:
                distribution.add_sample(s, d)
            expect(d.bucketCounts).to(equal(t['want']))
            _expect_stats_eq_direct_calc_from_samples(d, samples)

    def test_update_explict_buckets_ok(self):
        self.expect_adds_test_samples_ok(_make_explicit_dist)

    def test_update_exponential_buckets_ok(self):
        self.expect_adds_test_samples_ok(_make_exponential_dist)

    def test_update_linear_buckets_ok(self):
        self.expect_adds_test_samples_ok(_make_linear_dist)


class TestMerge(unittest2.TestCase):

    def setUp(self):
        self.merge_triples = (
            (
                distribution.create_exponential(3, 2, 0.1),
                distribution.create_exponential(3, 2, 0.1),
                distribution.create_exponential(4, 2, 0.1),
            ),(
                distribution.create_linear(3, 0.2, 0.1),
                distribution.create_linear(3, 0.2, 0.1),
                distribution.create_linear(4, 0.2, 0.1)
            ),(
                distribution.create_explicit([0.1, 0.3]),
                distribution.create_explicit([0.1, 0.3]),
                distribution.create_explicit([0.1, 0.3, 0.5]),
            )
        )
        for d1, d2, _ in self.merge_triples:
            distribution.add_sample(_LOW_SAMPLE, d1)
            distribution.add_sample(_HIGH_SAMPLE, d2)

    def test_should_fail_on_dissimilar_bucket_options(self):
        explicit = _make_explicit_dist()
        linear = _make_linear_dist()
        exponential = _make_exponential_dist()
        pairs = (
            (explicit, linear),
            (explicit, exponential),
            (linear, exponential)
        )
        for p in pairs:
            testf = lambda: distribution.merge(*p)
            expect(testf).to(raise_error(ValueError))

    def test_should_fail_on_dissimilar_bucket_counts(self):
        for _, d2, d3 in self.merge_triples:
            testf = lambda: distribution.merge(d2, d3)
            expect(testf).to(raise_error(ValueError))

    def test_should_merge_stats_correctly(self):
        # TODO(add a check of the variance)
        for d1, d2, _ in self.merge_triples:
            distribution.merge(d1, d2)
            expect(d2.count).to(equal(2))
            expect(d2.mean).to(equal((_HIGH_SAMPLE + _LOW_SAMPLE) / 2))
            expect(d2.maximum).to(equal(_HIGH_SAMPLE))
            expect(d2.minimum).to(equal(_LOW_SAMPLE))

    def test_should_merge_bucket_counts_correctly(self):
        for d1, d2, _ in self.merge_triples:
            d1_start = list(d1.bucketCounts)
            d2_start = list(d2.bucketCounts)
            want = [x + y for (x,y) in zip(d1_start, d2_start)]
            distribution.merge(d1, d2)
            expect(d2.bucketCounts).to(equal(want))



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import base64
import datetime
import unittest2
from expects import be_none, be_true, expect, equal, raise_error

from google.api.control import label_descriptor, messages, report_request

_KNOWN = label_descriptor.KnownLabels
ValueType = label_descriptor.ValueType

class KnownLabelsBase(object):
    SUBJECT = None
    GIVEN_INFO = report_request.Info(
        api_method = 'dummy_method',
        api_version = 'dummy_version',
        location = 'dummy_location',
        referer = 'dummy_referer'
    )
    WANTED_LABEL_DICT = {}

    def _matching_descriptor(self, hide_default=False):
        res = messages.LabelDescriptor(
            key=self.SUBJECT.label_name,
            valueType=self.SUBJECT.value_type)
        if res.valueType == ValueType.STRING and hide_default:
            res.valueType = None
        return res

    def _not_matched(self):
        d = self._matching_descriptor()
        d.valueType = ValueType.INT64  # no known labels have this type
        return d

    def test_should_be_supported(self):
        expect(_KNOWN.is_supported(self._matching_descriptor())).to(be_true)
        expect(_KNOWN.is_supported(
            self._matching_descriptor(hide_default=True))).to(be_true)
        expect(_KNOWN.is_supported(self._not_matched())).not_to(be_true)

    def test_should_be_matched_correctly(self):
        expect(self.SUBJECT.matches(self._matching_descriptor())).to(be_true)
        expect(self.SUBJECT.matches(
            self._matching_descriptor(hide_default=True))).to(be_true)
        expect(self.SUBJECT.matches(self._not_matched())).not_to(be_true)

    def test_should_update_request_info(self):
        given_dict = {}
        self.SUBJECT.do_labels_update(self.GIVEN_INFO, given_dict)
        expect(given_dict).to(equal(self.WANTED_LABEL_DICT))


class TestCredentialIdWithNoCreds(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.CREDENTIAL_ID


class TestCredentialIdWithApiKey(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.CREDENTIAL_ID
    GIVEN_INFO = report_request.Info(
        api_key = 'dummy_api_key',
    )
    WANTED_LABEL_DICT = {SUBJECT.label_name: 'apiKey:dummy_api_key'}


class TestCredentialIdWithAuthIssuer(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.CREDENTIAL_ID
    GIVEN_INFO = report_request.Info(
        auth_issuer = 'dummy_issuer',
        auth_audience = 'dummy_audience'
    )
    WANTED_VALUE = 'jwtAuth:issuer=' + base64.urlsafe_b64encode('dummy_issuer')
    WANTED_VALUE += '&audience=' + base64.urlsafe_b64encode('dummy_audience')
    WANTED_LABEL_DICT = {SUBJECT.label_name: WANTED_VALUE}


class EndUser(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.END_USER


class EndUserCountry(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.END_USER_COUNTRY


class ErrorType(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.ERROR_TYPE
    WANTED_LABEL_DICT = {SUBJECT.label_name: '2xx'}


class Protocol(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.PROTOCOL
    WANTED_LABEL_DICT = {
        SUBJECT.label_name: report_request.ReportedProtocols.UNKNOWN.name
    }


class Referer(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.REFERER
    WANTED_LABEL_DICT = {SUBJECT.label_name: 'dummy_referer'}


class ResponseCode(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.RESPONSE_CODE
    WANTED_LABEL_DICT = {SUBJECT.label_name: '200'}


class ResponseCodeClass(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.RESPONSE_CODE_CLASS
    WANTED_LABEL_DICT = {SUBJECT.label_name: '2xx'}


class StatusCodeWithOkStatus(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.STATUS_CODE
    WANTED_LABEL_DICT = {SUBJECT.label_name: '0'}


class StatusCodeWithKnown4XXStatus(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.STATUS_CODE
    GIVEN_INFO = report_request.Info(
        response_code = 401,
    )
    WANTED_LABEL_DICT = {SUBJECT.label_name: '16'}


class StatusCodeWithUnknown4XXStatus(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.STATUS_CODE
    GIVEN_INFO = report_request.Info(
        response_code = 477,
    )
    WANTED_LABEL_DICT = {SUBJECT.label_name: '9'}


class StatusCodeWithKnown5XXStatus(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.STATUS_CODE
    GIVEN_INFO = report_request.Info(
        response_code = 501,
    )
    WANTED_LABEL_DICT = {SUBJECT.label_name: '12'}


class StatusCodeWithUnknown5XXStatus(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.STATUS_CODE
    GIVEN_INFO = report_request.Info(
        response_code = 577,
    )
    WANTED_LABEL_DICT = {SUBJECT.label_name: '13'}


class StatusCodeWithUnknownStatus(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.STATUS_CODE
    GIVEN_INFO = report_request.Info(
        response_code = 777,
    )
    WANTED_LABEL_DICT = {SUBJECT.label_name: '2'}


class GaeCloneId(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.GAE_CLONE_ID


class GaeModuleId(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.GAE_MODULE_ID


class GaeReplicaIndex(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.GAE_REPLICA_INDEX


class GaeVersionId(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.GAE_VERSION_ID


class GcpLocation(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.GCP_LOCATION
    WANTED_LABEL_DICT = {SUBJECT.label_name: 'dummy_location'}


class GcpProject(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.GCP_PROJECT


class GcpRegion(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.GCP_REGION


class GcpResourceId(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.GCP_RESOURCE_ID


class GcpResourceType(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.GCP_RESOURCE_TYPE


class GcpService(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.GCP_SERVICE


class GcpZone(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.GCP_ZONE


class GcpUid(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.GCP_UID


class GcpApiMethod(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.GCP_API_METHOD
    WANTED_LABEL_DICT = {SUBJECT.label_name: 'dummy_method'}


class GcpApiVersion(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.GCP_API_VERSION
    WANTED_LABEL_DICT = {SUBJECT.label_name: 'dummy_version'}


class SccCallerIp(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.SCC_CALLER_IP


class SccPlatform(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.SCC_PLATFORM
    WANTED_LABEL_DICT = {
        SUBJECT.label_name: report_request.ReportedPlatforms.UNKNOWN.name
    }


class SccReferer(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.SCC_REFERER


class SccServiceAgent(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.SCC_SERVICE_AGENT
    WANTED_LABEL_DICT = {SUBJECT.label_name: label_descriptor.SERVICE_AGENT}


class SccUserAgent(KnownLabelsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.SCC_USER_AGENT
    WANTED_LABEL_DICT = {SUBJECT.label_name: label_descriptor.USER_AGENT}



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from apitools.base.py import encoding
import mock
import os
import tempfile
import unittest2
from expects import be_false, be_none, be_true, expect, equal, raise_error

from google.api.auth import suppliers
from google.api.auth import tokens
from google.api.control import client, messages, report_request, service, wsgi


def _dummy_start_response(content, dummy_response_headers):
    pass


_DUMMY_RESPONSE = ('All must answer "here!"',)


class _DummyWsgiApp(object):

    def __call__(self, environ, dummy_start_response):
        return _DUMMY_RESPONSE


class TestEnvironmentMiddleware(unittest2.TestCase):

    def test_should_add_service_et_al_to_environment(self):
        cls = wsgi.EnvironmentMiddleware
        wrappee = _DummyWsgiApp()
        wanted_service = service.Loaders.SIMPLE.load()
        wrapped = cls(wrappee, wanted_service)

        given = {
            'wsgi.url_scheme': 'http',
            'HTTP_HOST': 'localhost',
            'REQUEST_METHOD': 'GET'
        }
        wrapped(given, _dummy_start_response)
        expect(given.get(cls.SERVICE)).to(equal(wanted_service))
        expect(given.get(cls.SERVICE_NAME)).to(equal(wanted_service.name))
        expect(given.get(cls.METHOD_REGISTRY)).not_to(be_none)
        expect(given.get(cls.REPORTING_RULES)).not_to(be_none)
        expect(given.get(cls.METHOD_INFO)).not_to(be_none)


class TestMiddleware(unittest2.TestCase):
    PROJECT_ID = 'middleware'

    def test_should_not_send_requests_if_there_is_no_service(self):
        wrappee = _DummyWsgiApp()
        control_client = mock.MagicMock(spec=client.Client)

        given = {
            'wsgi.url_scheme': 'http',
            'PATH_INFO': '/any/method',
            'REMOTE_ADDR': '192.168.0.3',
            'HTTP_HOST': 'localhost',
            'HTTP_REFERER': 'example.myreferer.com',
            'REQUEST_METHOD': 'GET'
        }
        dummy_response = messages.CheckResponse(operationId='fake_operation_id')
        wrapped = wsgi.Middleware(wrappee, self.PROJECT_ID, control_client)
        wrapped(given, _dummy_start_response)
        expect(control_client.check.called).to(be_false)
        expect(control_client.report.called).to(be_false)

    def test_should_send_requests_using_the_client(self):
        wrappee = _DummyWsgiApp()
        control_client = mock.MagicMock(spec=client.Client)

        given = {
            'wsgi.url_scheme': 'http',
            'PATH_INFO': '/any/method',
            'REMOTE_ADDR': '192.168.0.3',
            'HTTP_HOST': 'localhost',
            'HTTP_REFERER': 'example.myreferer.com',
            'REQUEST_METHOD': 'GET'
        }
        dummy_response = messages.CheckResponse(operationId='fake_operation_id')
        with_control = wsgi.Middleware(wrappee, self.PROJECT_ID, control_client)
        wrapped = wsgi.EnvironmentMiddleware(with_control,
                                             service.Loaders.SIMPLE.load())
        control_client.check.return_value = dummy_response
        wrapped(given, _dummy_start_response)
        expect(control_client.check.called).to(be_true)
        expect(control_client.report.called).to(be_true)

    def test_should_send_report_request_if_check_fails(self):
        wrappee = _DummyWsgiApp()
        control_client = mock.MagicMock(spec=client.Client)
        given = {
            'wsgi.url_scheme': 'http',
            'PATH_INFO': '/any/method',
            'REMOTE_ADDR': '192.168.0.3',
            'HTTP_HOST': 'localhost',
            'HTTP_REFERER': 'example.myreferer.com',
            'REQUEST_METHOD': 'GET'
        }
        dummy_response = messages.CheckResponse(
            operationId = 'fake_operation_id',
            checkErrors = [
                messages.CheckError(
                    code=messages.CheckError.CodeValueValuesEnum.PROJECT_DELETED)
            ]
        )
        wrapped = wsgi.add_all(wrappee,
                               self.PROJECT_ID,
                               control_client,
                               loader=service.Loaders.SIMPLE)
        control_client.check.return_value = dummy_response
        wrapped(given, _dummy_start_response)
        expect(control_client.check.called).to(be_true)
        expect(control_client.report.called).to(be_true)

    def test_load_service_failed(self):
        loader = mock.MagicMock(load=lambda: None)
        with self.assertRaisesRegex(ValueError, "Failed to load service config"):
            wsgi.add_all(_DummyWsgiApp(),
                         self.PROJECT_ID,
                         mock.MagicMock(spec=client.Client),
                         loader=loader)


_SYSTEM_PARAMETER_CONFIG_TEST = """
{
    "name": "system-parameter-config",
    "systemParameters": {
       "rules": [{
         "selector": "Uvw.Method1",
         "parameters": [{
            "name": "name1",
            "httpHeader": "Header-Key1",
            "urlQueryParameter": "param_key1"
         }, {
            "name": "name2",
            "httpHeader": "Header-Key2",
            "urlQueryParameter": "param_key2"
         }, {
            "name": "api_key",
            "httpHeader": "ApiKeyHeader",
            "urlQueryParameter": "ApiKeyParam"
         }, {
            "httpHeader": "Ignored-NoName-Key3",
            "urlQueryParameter": "Ignored-NoName-key3"
         }]
       }, {
         "selector": "Bad.NotConfigured",
         "parameters": [{
            "name": "neverUsed",
            "httpHeader": "NeverUsed-Key1",
            "urlQueryParameter": "NeverUsed_key1"
         }]
       }]
    },
    "http": {
        "rules": [{
            "selector": "Uvw.Method1",
            "get": "/uvw/method1/*"
        }, {
            "selector": "Uvw.MethodNeedsApiKey",
            "get": "/uvw/method_needs_api_key/*"
        }, {
            "selector": "Uvw.DefaultParameters",
            "get": "/uvw/default_parameters"
        }]
    },
    "usage": {
        "rules": [{
            "selector" : "Uvw.Method1",
            "allowUnregisteredCalls" : true
        },  {
            "selector": "Uvw.MethodNeedsApiKey",
            "allowUnregisteredCalls" : false
        }, {
            "selector" : "Uvw.DefaultParameters",
            "allowUnregisteredCalls" : true
        }]
    }
}
"""

class TestMiddlewareWithParams(unittest2.TestCase):
    PROJECT_ID = 'middleware-with-params'

    def setUp(self):
        _config_fd = tempfile.NamedTemporaryFile(delete=False)
        with _config_fd as f:
            f.write(_SYSTEM_PARAMETER_CONFIG_TEST)
        self._config_file = _config_fd.name
        os.environ[service.CONFIG_VAR] = self._config_file

    def tearDown(self):
        if os.path.exists(self._config_file):
            os.remove(self._config_file)

    def test_should_send_requests_with_no_param(self):
        wrappee = _DummyWsgiApp()
        control_client = mock.MagicMock(spec=client.Client)
        given = {
            'wsgi.url_scheme': 'http',
            'PATH_INFO': '/uvw/method1/with_no_param',
            'REMOTE_ADDR': '192.168.0.3',
            'HTTP_HOST': 'localhost',
            'HTTP_REFERER': 'example.myreferer.com',
            'REQUEST_METHOD': 'GET'
        }
        dummy_response = messages.CheckResponse(operationId='fake_operation_id')
        wrapped = wsgi.add_all(wrappee,
                               self.PROJECT_ID,
                               control_client,
                               loader=service.Loaders.ENVIRONMENT)
        control_client.check.return_value = dummy_response
        wrapped(given, _dummy_start_response)
        expect(control_client.check.called).to(be_true)
        req = control_client.check.call_args[0][0]
        expect(req.checkRequest.operation.consumerId).to(
            equal('project:middleware-with-params'))
        expect(control_client.report.called).to(be_true)

    def test_should_send_requests_with_configured_query_param_api_key(self):
        wrappee = _DummyWsgiApp()
        control_client = mock.MagicMock(spec=client.Client)
        given = {
            'wsgi.url_scheme': 'http',
            'QUERY_STRING': 'ApiKeyParam=my-query-value',
            'PATH_INFO': '/uvw/method1/with_query_param',
            'REMOTE_ADDR': '192.168.0.3',
            'HTTP_HOST': 'localhost',
            'HTTP_REFERER': 'example.myreferer.com',
            'REQUEST_METHOD': 'GET'
        }
        dummy_response = messages.CheckResponse(operationId='fake_operation_id')
        wrapped = wsgi.add_all(wrappee,
                               self.PROJECT_ID,
                               control_client,
                               loader=service.Loaders.ENVIRONMENT)
        control_client.check.return_value = dummy_response
        wrapped(given, _dummy_start_response)
        expect(control_client.check.called).to(be_true)
        check_req = control_client.check.call_args[0][0]
        expect(check_req.checkRequest.operation.consumerId).to(
            equal('api_key:my-query-value'))
        expect(control_client.report.called).to(be_true)
        report_req = control_client.report.call_args[0][0]
        expect(report_req.reportRequest.operations[0].consumerId).to(
            equal('api_key:my-query-value'))

    def test_should_send_requests_with_configured_header_api_key(self):
        wrappee = _DummyWsgiApp()
        control_client = mock.MagicMock(spec=client.Client)
        given = {
            'wsgi.url_scheme': 'http',
            'PATH_INFO': '/uvw/method1/with_query_param',
            'REMOTE_ADDR': '192.168.0.3',
            'HTTP_HOST': 'localhost',
            'HTTP_APIKEYHEADER': 'my-header-value',
            'HTTP_REFERER': 'example.myreferer.com',
            'REQUEST_METHOD': 'GET'
        }
        dummy_response = messages.CheckResponse(operationId='fake_operation_id')
        wrapped = wsgi.add_all(wrappee,
                               self.PROJECT_ID,
                               control_client,
                               loader=service.Loaders.ENVIRONMENT)
        control_client.check.return_value = dummy_response
        wrapped(given, _dummy_start_response)
        expect(control_client.check.called).to(be_true)
        check_request = control_client.check.call_args_list[0].checkRequest
        check_req = control_client.check.call_args[0][0]
        expect(check_req.checkRequest.operation.consumerId).to(
            equal('api_key:my-header-value'))
        expect(control_client.report.called).to(be_true)
        report_req = control_client.report.call_args[0][0]
        expect(report_req.reportRequest.operations[0].consumerId).to(
            equal('api_key:my-header-value'))

    def test_should_send_requests_with_default_query_param_api_key(self):
        for default_key in ('key', 'api_key'):
            wrappee = _DummyWsgiApp()
            control_client = mock.MagicMock(spec=client.Client)
            given = {
                'wsgi.url_scheme': 'http',
                'QUERY_STRING': '%s=my-default-api-key-value' % (default_key,),
                'PATH_INFO': '/uvw/method_needs_api_key/with_query_param',
                'REMOTE_ADDR': '192.168.0.3',
                'HTTP_HOST': 'localhost',
                'HTTP_REFERER': 'example.myreferer.com',
                'REQUEST_METHOD': 'GET'
            }
            dummy_response = messages.CheckResponse(operationId='fake_operation_id')
            wrapped = wsgi.add_all(wrappee,
                                   self.PROJECT_ID,
                                   control_client,
                                   loader=service.Loaders.ENVIRONMENT)
            control_client.check.return_value = dummy_response
            wrapped(given, _dummy_start_response)
            expect(control_client.check.called).to(be_true)
            check_request = control_client.check.call_args_list[0].checkRequest
            check_req = control_client.check.call_args[0][0]
            expect(check_req.checkRequest.operation.consumerId).to(
                equal('api_key:my-default-api-key-value'))
            expect(control_client.report.called).to(be_true)
            report_req = control_client.report.call_args[0][0]
            expect(report_req.reportRequest.operations[0].consumerId).to(
                equal('api_key:my-default-api-key-value'))

    def test_should_not_perform_check_if_needed_api_key_is_missing(self):
        wrappee = _DummyWsgiApp()
        control_client = mock.MagicMock(spec=client.Client)
        given = {
            'wsgi.url_scheme': 'http',
            'PATH_INFO': '/uvw/method_needs_api_key/more_stuff',
            'REMOTE_ADDR': '192.168.0.3',
            'HTTP_HOST': 'localhost',
            'HTTP_REFERER': 'example.myreferer.com',
            'REQUEST_METHOD': 'GET'
        }
        dummy_response = messages.CheckResponse(operationId='fake_operation_id')
        wrapped = wsgi.add_all(wrappee,
                               self.PROJECT_ID,
                               control_client,
                               loader=service.Loaders.ENVIRONMENT)
        control_client.check.return_value = dummy_response
        wrapped(given, _dummy_start_response)
        expect(control_client.check.called).to(be_false)
        expect(control_client.report.called).to(be_true)
        report_req = control_client.report.call_args[0][0]
        expect(report_req.reportRequest.operations[0].consumerId).to(
            equal('project:middleware-with-params'))

AuthMiddleware = wsgi.AuthenticationMiddleware


class TestAuthenticationMiddleware(unittest2.TestCase):

  def setUp(self):
      self._mock_application = _DummyWsgiApp()
      self._mock_authenticator = mock.MagicMock(spec=tokens.Authenticator)
      self._middleware = AuthMiddleware(self._mock_application,
                                        self._mock_authenticator)

  def test_no_authentication(self):
      with self.assertRaisesRegex(ValueError, "Invalid authenticator"):
          AuthMiddleware(self._mock_application, None)

  def test_no_method_info(self):
      environ = {}
      self.assertEqual(_DUMMY_RESPONSE,
                       self._middleware(environ, _dummy_start_response))

  def test_no_auth_token(self):
      auth_app = AuthMiddleware(self.UserInfoWsgiApp(), self._mock_authenticator)
      method_info = mock.MagicMock()
      method_info.auth_info = mock.MagicMock()
      environ = {
          wsgi.EnvironmentMiddleware.METHOD_INFO: method_info
      }
      self.assertIsNone(auth_app(environ, _dummy_start_response))

  def test_malformed_authorization_header(self):
      auth_app = AuthMiddleware(self.UserInfoWsgiApp(), self._mock_authenticator)
      environ = {
          "HTTP_AUTHORIZATION": "malformed-auth-token",
          wsgi.EnvironmentMiddleware.METHOD_INFO: mock.MagicMock(),
          wsgi.EnvironmentMiddleware.SERVICE_NAME: "service-name"
      }
      self._mock_authenticator.authenticate.side_effect = suppliers.UnauthenticatedException()
      self.assertIsNone(auth_app(environ, _dummy_start_response))

  def test_successful_authentication(self):
      auth_token = "Bearer test-bearer-token"
      auth_info = mock.MagicMock()
      service_name = "test-service-name"
      method_info = mock.MagicMock()
      method_info.auth_info = auth_info
      environ = {
          "HTTP_AUTHORIZATION": auth_token,
          wsgi.EnvironmentMiddleware.METHOD_INFO: method_info,
          wsgi.EnvironmentMiddleware.SERVICE_NAME: service_name
      }

      user_info = mock.MagicMock()
      self._mock_authenticator.authenticate.return_value = user_info
      self._middleware(environ, _dummy_start_response)
      self.assertEqual(user_info, environ.get(AuthMiddleware.USER_INFO))
      authenticate_mock = self._mock_authenticator.authenticate
      authenticate_mock.assert_called_once_with("test-bearer-token", auth_info,
                                                service_name)

  def test_auth_token_in_query(self):
      auth_token = "test-bearer-token"
      auth_info = mock.MagicMock()
      service_name = "test-service-name"
      method_info = mock.MagicMock()
      method_info.auth_info = auth_info
      environ = {
          "QUERY_STRING": "access_token=" + auth_token,
          wsgi.EnvironmentMiddleware.METHOD_INFO: method_info,
          wsgi.EnvironmentMiddleware.SERVICE_NAME: service_name
      }

      user_info = mock.MagicMock()
      self._mock_authenticator.authenticate.return_value = user_info
      self._middleware(environ, _dummy_start_response)
      self.assertEqual(user_info, environ.get(AuthMiddleware.USER_INFO))
      authenticate_mock = self._mock_authenticator.authenticate
      authenticate_mock.assert_called_once_with("test-bearer-token", auth_info,
                                                service_name)

  patched_environ = {}
  @mock.patch("os.environ", patched_environ)
  def test_set_user_info(self):
    environ = {
        "QUERY_STRING": "access_token=test-token",
        wsgi.EnvironmentMiddleware.METHOD_INFO: mock.MagicMock(),
        wsgi.EnvironmentMiddleware.SERVICE_NAME: "test-service-name"
    }
    application = self.UserInfoWsgiApp()
    auth_middleware = AuthMiddleware(application, self._mock_authenticator)
    user_info = mock.MagicMock()
    self._mock_authenticator.authenticate.return_value = user_info
    self.assertEqual(user_info, auth_middleware(environ,
                                                _dummy_start_response))
    self.assertFalse(self.patched_environ)

  class UserInfoWsgiApp(object):
    def __call__(self, environ, start_response):
      return os.environ.get(wsgi.AuthenticationMiddleware.USER_INFO)


class TestCreateAuthenticator(unittest2.TestCase):
    def test_create_without_service(self):
        with self.assertRaises(ValueError):
            wsgi._create_authenticator(None)

    def test_load_service_without_auth(self):
        service = _read_service_from_json("{}")
        self.assertIsNone(wsgi._create_authenticator(service))

    def test_load_service(self):
        json = """{
            "authentication": {
                "providers": [{
                    "issuer": "auth-issuer"
                }]
            }
        }"""
        service = _read_service_from_json(json)
        self.assertIsNotNone(wsgi._create_authenticator(service))


patched_platform_environ = {}
@mock.patch.dict('os.environ', patched_platform_environ, clear=True)
class TestPlatformDetection(unittest2.TestCase):

  def test_development(self):
    os.environ['SERVER_SOFTWARE'] = 'Development/2.0.0'
    self.assertEqual(report_request.ReportedPlatforms.DEVELOPMENT,
                     wsgi._get_platform())

  def test_gke(self):
    os.environ['KUBERNETES_SERVICE_HOST'] = 'hostname'
    self.assertEqual(report_request.ReportedPlatforms.GKE,
                     wsgi._get_platform())

  @mock.patch.object(wsgi, '_running_on_gce', return_value=True)
  def test_gae_flex(self, _running_on_gce):
    os.environ['GAE_MODULE_NAME'] = 'gae_module'
    self.assertEqual(report_request.ReportedPlatforms.GAE_FLEX,
                     wsgi._get_platform())

  @mock.patch.object(wsgi, '_running_on_gce', return_value=True)
  def test_gce(self, _running_on_gce):
    self.assertEqual(report_request.ReportedPlatforms.GCE,
                     wsgi._get_platform())

  @mock.patch.object(wsgi, '_running_on_gce', return_value=False)
  def test_gae_standard(self, _running_on_gce):
    os.environ['GAE_MODULE_NAME'] = 'gae_module'
    self.assertEqual(report_request.ReportedPlatforms.GAE_STANDARD,
                     wsgi._get_platform())

  @mock.patch.object(wsgi, '_running_on_gce', return_value=False)
  def test_unknown(self, _running_on_gce):
    self.assertEqual(report_request.ReportedPlatforms.UNKNOWN,
                     wsgi._get_platform())


def _read_service_from_json(json):
    return encoding.JsonToMessage(messages.Service, json)



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides a utility method that generates auth token."""

import json

from jwkest import jws


def generate_auth_token(payload, keys, alg="ES256", kid=None):
  json_web_signature = jws.JWS(json.dumps(payload), alg=alg, kid=kid)
  return json_web_signature.sign_compact(keys=keys)



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest
import httmock
import mock

from Crypto import PublicKey
from jwkest import jwk

from google.api.auth import suppliers


class KeyUriSupplierTest(unittest.TestCase):

  def test_supply_issuer(self):
    issuer = "https://issuer.com"
    jwks_uri = "https://issuer.com/jwks/uri"
    configs = {issuer: suppliers.IssuerUriConfig(False, jwks_uri)}
    supplier = suppliers.KeyUriSupplier(configs)
    self.assertEquals(jwks_uri, supplier.supply(issuer))
    self.assertIsNone(supplier.supply("random-issuer"))

  def test_openid_discovery(self):
    jwks_uri = "https://issuer.com/jwks/uri"
    @httmock.urlmatch(scheme="https", netloc="issuer.com",
                      path="/" + suppliers._OPEN_ID_CONFIG_PATH)
    def _mock_response(url, request):  # pylint: disable=unused-argument
      response = {"jwks_uri": jwks_uri}
      return json.dumps(response)

    issuer = "https://issuer.com"
    configs = {issuer: suppliers.IssuerUriConfig(True, None)}
    supplier = suppliers.KeyUriSupplier(configs)
    with httmock.HTTMock(_mock_response):
      self.assertEquals(jwks_uri, supplier.supply(issuer))

  def test_issuer_without_protocol(self):
    jwks_uri = "https://issuer.com/jwks/uri"
    @httmock.urlmatch(scheme="https", netloc="issuer.com",
                      path="/" + suppliers._OPEN_ID_CONFIG_PATH)
    def _mock_response(url, request):  # pylint: disable=unused-argument
      response = {"jwks_uri": jwks_uri}
      return json.dumps(response)

    # Specify an issuer without protocol to make sure the "https://" prefix is
    # added automatically.
    issuer = "issuer.com"
    configs = {issuer: suppliers.IssuerUriConfig(True, None)}
    supplier = suppliers.KeyUriSupplier(configs)
    with httmock.HTTMock(_mock_response):
      self.assertEquals(jwks_uri, supplier.supply(issuer))

  def test_openid_discovery_with_bad_json(self):
    @httmock.urlmatch(scheme="https", netloc="issuer.com")
    def _mock_response_with_bad_json(url, request):  # pylint: disable=unused-argument
      return "bad-json"

    issuer = "https://issuer.com"
    configs = {issuer: suppliers.IssuerUriConfig(True, None)}
    supplier = suppliers.KeyUriSupplier(configs)
    with httmock.HTTMock(_mock_response_with_bad_json):
      with self.assertRaises(suppliers.UnauthenticatedException):
        supplier.supply(issuer)


class JwksSupplierTest(unittest.TestCase):
  _mock_timer = mock.MagicMock()

  def setUp(self):
    self._key_uri_supplier = mock.MagicMock()
    self._jwks_uri_supplier = suppliers.JwksSupplier(self._key_uri_supplier)

  def test_supply_with_unknown_issuer(self):
    self._key_uri_supplier.supply.return_value = None
    issuer = "unknown-issuer"
    expected_message = "Cannot find the `jwks_uri` for issuer " + issuer
    with self.assertRaisesRegexp(suppliers.UnauthenticatedException,
                                 expected_message):
      self._jwks_uri_supplier.supply(issuer)

  def test_supply_with_invalid_json_response(self):
    scheme = "https"
    issuer = "issuer.com"
    self._key_uri_supplier.supply.return_value = scheme + "://" + issuer

    @httmock.urlmatch(scheme=scheme, netloc=issuer)
    def _mock_response_with_invalid_json(url, response):  # pylint: disable=unused-argument
      return "invalid-json"

    with httmock.HTTMock(_mock_response_with_invalid_json):
      with self.assertRaises(suppliers.UnauthenticatedException):
        self._jwks_uri_supplier.supply(issuer)

  def test_supply_jwks(self):
    rsa_key = PublicKey.RSA.generate(2048)
    jwks = jwk.KEYS()
    jwks.wrap_add(rsa_key)

    scheme = "https"
    issuer = "issuer.com"
    self._key_uri_supplier.supply.return_value = scheme + "://" + issuer

    @httmock.urlmatch(scheme=scheme, netloc=issuer)
    def _mock_response_with_jwks(url, response):  # pylint: disable=unused-argument
      return jwks.dump_jwks()

    with httmock.HTTMock(_mock_response_with_jwks):
      actual_jwks = self._jwks_uri_supplier.supply(issuer)
      self.assertEquals(1, len(actual_jwks))
      actual_key = actual_jwks[0].key
      self.assertEquals(rsa_key.n, actual_key.n)
      self.assertEquals(rsa_key.e, actual_key.e)

  def test_supply_jwks_with_x509_certificate(self):
    rsa_key = PublicKey.RSA.generate(2048)
    cert = rsa_key.publickey().exportKey("PEM")
    kid = "rsa-cert"

    scheme = "https"
    issuer = "issuer.com"
    self._key_uri_supplier.supply.return_value = scheme + "://" + issuer

    @httmock.urlmatch(scheme=scheme, netloc=issuer)
    def _mock_response_with_x509_certificates(url, response):  # pylint: disable=unused-argument
      return json.dumps({kid: cert})

    with httmock.HTTMock(_mock_response_with_x509_certificates):
      actual_jwks = self._jwks_uri_supplier.supply(issuer)
      self.assertEquals(1, len(actual_jwks))
      actual_key = actual_jwks[0].key

      self.assertEquals(kid, actual_jwks[0].kid)
      self.assertEquals(rsa_key.n, actual_key.n)
      self.assertEquals(rsa_key.e, actual_key.e)

  def test_supply_empty_x509_certificate(self):
    scheme = "https"
    issuer = "issuer.com"
    self._key_uri_supplier.supply.return_value = scheme + "://" + issuer

    @httmock.urlmatch(scheme=scheme, netloc=issuer)
    def _mock_invalid_response(url, response):  # pylint: disable=unused-argument
      return json.dumps({"kid": "invlid-certificate"})

    with httmock.HTTMock(_mock_invalid_response):
      with self.assertRaises(suppliers.UnauthenticatedException):
        self._jwks_uri_supplier.supply(issuer)

  @mock.patch("time.time", _mock_timer)
  def test_supply_cached_jwks(self):
    rsa_key = PublicKey.RSA.generate(2048)
    jwks = jwk.KEYS()
    jwks.wrap_add(rsa_key)

    scheme = "https"
    issuer = "issuer.com"
    self._key_uri_supplier.supply.return_value = scheme + "://" + issuer

    @httmock.urlmatch(scheme=scheme, netloc=issuer)
    def _mock_response_with_jwks(url, response):  # pylint: disable=unused-argument
      return jwks.dump_jwks()

    with httmock.HTTMock(_mock_response_with_jwks):
      JwksSupplierTest._mock_timer.return_value = 10
      self.assertEqual(1, len(self._jwks_uri_supplier.supply(issuer)))

      # Add an additional key to the JWKS to be returned by the HTTP request.
      jwks.wrap_add(PublicKey.RSA.generate(2048))

      # Forward the clock by 1 second. The JWKS should remain cached.
      JwksSupplierTest._mock_timer.return_value += 1
      self._jwks_uri_supplier.supply(issuer)
      self.assertEqual(1, len(self._jwks_uri_supplier.supply(issuer)))

      # Forward the clock by 5 minutes. The cache entry should have expired so
      # the returned JWKS should be the updated one with two keys.
      JwksSupplierTest._mock_timer.return_value += 5 * 60
      self._jwks_uri_supplier.supply(issuer)
      self.assertEqual(2, len(self._jwks_uri_supplier.supply(issuer)))



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""check_request supports aggregation of CheckRequests.

:func:`sign` generated a signature from CheckRequests
:class:`~google.api.gen.servicecontrol_v1_message.Operation` represents
information regarding an operation, and is a key constituent of
:class:`~google.api.gen.servicecontrol_v1_message.CheckRequest` and
:class:`~google.api.gen.servicecontrol_v1_message.ReportRequests.

The :class:`.Aggregator` implements the strategy for aggregating CheckRequests
and caching their responses.

"""

from __future__ import absolute_import

import collections
import hashlib
import httplib
import logging
from datetime import datetime

from apitools.base.py import encoding

from . import caches, label_descriptor, messages
from . import metric_value, operation, signing

logger = logging.getLogger(__name__)

# alias for brevity
_CheckErrors = messages.CheckError.CodeValueValuesEnum
_IS_OK = (httplib.OK, '', True)
_IS_UNKNOWN = (
    httplib.INTERNAL_SERVER_ERROR,
    'Request blocked due to unsupported block reason {detail}',
    False)
_CHECK_ERROR_CONVERSION = {
    _CheckErrors.NOT_FOUND: (
        httplib.BAD_REQUEST,
        'Client project not found. Please pass a valid project',
        False,
    ),
    _CheckErrors.API_KEY_NOT_FOUND: (
        httplib.BAD_REQUEST,
        'API key not found. Please pass a valid API key',
        True,
    ),
    _CheckErrors.API_KEY_EXPIRED: (
        httplib.BAD_REQUEST,
        'API key expired. Please renew the API key',
        True,
    ),
    _CheckErrors.API_KEY_INVALID: (
        httplib.BAD_REQUEST,
        'API not valid. Please pass a valid API key',
        True,
    ),
    _CheckErrors.SERVICE_NOT_ACTIVATED: (
        httplib.FORBIDDEN,
        '{detail} Please enable the project for {project_id}',
        False,
    ),
    _CheckErrors.PERMISSION_DENIED: (
        httplib.FORBIDDEN,
        'Permission denied: {detail}',
        False,
    ),
    _CheckErrors.IP_ADDRESS_BLOCKED: (
        httplib.FORBIDDEN,
        '{detail}',
        False,
    ),
    _CheckErrors.REFERER_BLOCKED: (
        httplib.FORBIDDEN,
        '{detail}',
        False,
    ),
    _CheckErrors.CLIENT_APP_BLOCKED: (
        httplib.FORBIDDEN,
        '{detail}',
        False,
    ),
    _CheckErrors.PROJECT_DELETED: (
        httplib.FORBIDDEN,
        'Project {project_id} has been deleted',
        False,
    ),
    _CheckErrors.PROJECT_INVALID: (
        httplib.BAD_REQUEST,
        'Client Project is not valid.  Please pass a valid project',
        False,
    ),
    _CheckErrors.VISIBILITY_DENIED: (
        httplib.FORBIDDEN,
        'Project {project_id} has no visibility access to the service',
        False,
    ),
    _CheckErrors.BILLING_DISABLED: (
        httplib.FORBIDDEN,
        'Project {project_id} has billing disabled. Please enable it',
        False,
    ),

    # Fail open for internal server errors
    _CheckErrors.NAMESPACE_LOOKUP_UNAVAILABLE: _IS_OK,
    _CheckErrors.SERVICE_STATUS_UNAVAILABLE: _IS_OK,
    _CheckErrors.BILLING_STATUS_UNAVAILABLE: _IS_OK,
    _CheckErrors.QUOTA_CHECK_UNAVAILABLE: _IS_OK,
}


def convert_response(check_response, project_id):
    """Computes a http status code and message `CheckResponse`

    The return value a tuple (code, message, api_key_is_bad) where

    code: is the http status code
    message: is the message to return
    api_key_is_bad: indicates that a given api_key is bad

    Args:
       check_response (:class:`google.api.gen.servicecontrol_v1_messages.CheckResponse`):
         the response from calling an api

    Returns:
       tuple(code, message, bool)
    """
    if not check_response or not check_response.checkErrors:
        return _IS_OK

    # only check the first error for now, as per ESP
    theError = check_response.checkErrors[0]
    error_tuple = _CHECK_ERROR_CONVERSION.get(theError.code, _IS_UNKNOWN)
    if error_tuple[1].find('{') == -1:  # no replacements needed:
        return error_tuple

    updated_msg = error_tuple[1].replace('{project_id}', project_id)
    updated_msg = updated_msg.replace('{detail}', theError.detail or '')
    error_tuple = (error_tuple[0], updated_msg, error_tuple[2])
    return error_tuple


def sign(check_request):
    """Obtains a signature for an operation in a `CheckRequest`

    Args:
       op (:class:`google.api.gen.servicecontrol_v1_messages.Operation`): an
         operation used in a `CheckRequest`

    Returns:
       string: a secure hash generated from the operation
    """
    if not isinstance(check_request, messages.CheckRequest):
        raise ValueError('Invalid request')
    op = check_request.operation
    if op is None or op.operationName is None or op.consumerId is None:
        logging.error('Bad %s: not initialized => not signed', check_request)
        raise ValueError('check request must be initialized with an operation')
    md5 = hashlib.md5()
    md5.update(op.operationName)
    md5.update('\x00')
    md5.update(op.consumerId)
    if op.labels:
        signing.add_dict_to_hash(md5, encoding.MessageToPyValue(op.labels))
    for value_set in op.metricValueSets:
        md5.update('\x00')
        md5.update(value_set.metricName)
        for mv in value_set.metricValues:
            metric_value.update_hash(md5, mv)

    md5.update('\x00')
    if op.quotaProperties:
        # N.B: this differs form cxx implementation, which serializes the
        # protobuf. This should be OK as the exact hash used does not need to
        # match across implementations.
        md5.update(repr(op.quotaProperties))

    md5.update('\x00')
    return md5.digest()


_KNOWN_LABELS = label_descriptor.KnownLabels


class Info(collections.namedtuple('Info',
                                  ('client_ip',) + operation.Info._fields),
           operation.Info):
    """Holds the information necessary to fill in CheckRequest.

    In addition the attributes in :class:`operation.Info`, this has:

    Attributes:
       client_ip: the client IP address

    """
    def __new__(cls, client_ip='', **kw):
        """Invokes the base constructor with default values."""
        op_info = operation.Info(**kw)
        return super(Info, cls).__new__(cls, client_ip, **op_info._asdict())

    def as_check_request(self, timer=datetime.utcnow):
        """Makes a `ServicecontrolServicesCheckRequest` from this instance

        Returns:
          a ``ServicecontrolServicesCheckRequest``

        Raises:
          ValueError: if the fields in this instance are insufficient to
            to create a valid ``ServicecontrolServicesCheckRequest``

        """
        if not self.service_name:
            raise ValueError('the service name must be set')
        if not self.operation_id:
            raise ValueError('the operation id must be set')
        if not self.operation_name:
            raise ValueError('the operation name must be set')
        op = super(Info, self).as_operation(timer=timer)
        labels = {
            _KNOWN_LABELS.SCC_USER_AGENT.label_name: label_descriptor.USER_AGENT
        }
        if self.client_ip:
            labels[_KNOWN_LABELS.SCC_CALLER_IP.label_name] = self.client_ip

        if self.referer:
            labels[_KNOWN_LABELS.SCC_REFERER.label_name] = self.referer

        op.labels = encoding.PyValueToMessage(
            messages.Operation.LabelsValue, labels)
        check_request = messages.CheckRequest(operation=op)
        return messages.ServicecontrolServicesCheckRequest(
            serviceName=self.service_name,
            checkRequest=check_request)


class Aggregator(object):
    """Caches and aggregates ``CheckRequests``.

    Concurrency: Thread safe.

    Usage:

    Creating a new cache entry and use cached response

    Example:
      >>> options = caches.CheckOptions()
      >>> agg = Aggregator('my_service', options)
      >>> req = ServicecontrolServicesCheckRequest(...)
      >>> # check returns None as the request is not cached
      >>> if agg.check(req) is not None:
      ...    resp = service.check(req)
      ...    agg = service.add_response(req, resp)
      >>> agg.check(req)  # response now cached according as-per options
      <CheckResponse ....>

    Refreshing a cached entry after a flush interval

    The flush interval is constrained to be shorter than the actual cache
    expiration.  This allows the response to potentially remain cached and be
    aggregated with subsequent check requests for the same operation.

    Example:
      >>> # continuing from the previous example,
      >>> # ... after the flush interval
      >>> # - the response is still in the cache, i.e, not expired
      >>> # - the first call after the flush interval returns None, subsequent
      >>> #  calls continue to return the cached response
      >>> agg.check(req)  # signals the caller to call service.check(req)
      None
      >>> agg.check(req)  # next call returns the cached response
      <CheckResponse ....>

    Flushing the cache

    Once a response is expired, if there is an outstanding, cached CheckRequest
    for it, this should be sent and their responses added back to the
    aggregator instance, as they will contain quota updates that have not been
    sent.

    Example:

      >>> # continuing the previous example
      >>> for req in agg.flush():  # an iterable of cached CheckRequests
      ...     resp = caller.send_req(req)  # caller sends them
      >>>     agg.add_response(req, resp)  # and caches their responses

    """

    def __init__(self, service_name, options, kinds=None,
                 timer=datetime.utcnow):
        """Constructor.

        Args:
          service_name (string): names the service that all requests aggregated
            by this instance will be sent
          options (:class:`~google.api.caches.CheckOptions`): configures the
            caching and flushing behavior of this instance
          kinds (dict[string,[google.api.control.MetricKind]]): specifies the
            kind of metric for each each metric name.
          timer (function([[datetime]]): a function that returns the current
            as a time as a datetime instance
        """
        self._service_name = service_name
        self._options = options
        self._cache = caches.create(options, timer=timer)
        self._kinds = {} if kinds is None else dict(kinds)
        self._timer = timer

    @property
    def service_name(self):
        """The service to which all aggregated requests should belong."""
        return self._service_name

    @property
    def flush_interval(self):
        """The interval between calls to flush.

        Returns:
           timedelta: the period between calls to flush if, or ``None`` if no
           cache is set

        """
        return None if self._cache is None else self._options.expiration

    def flush(self):
        """Flushes this instance's cache.

        The driver of this instance should call this method every
        `flush_interval`.

        Returns:
          list['CheckRequest']: corresponding to CheckRequests that were
          pending

        """
        if self._cache is None:
            return []
        with self._cache as c:
            flushed_items = list(c.out_deque)
            c.out_deque.clear()
            cached_reqs = [item.extract_request() for item in flushed_items]
            cached_reqs = [req for req in cached_reqs if req is not None]
            return cached_reqs

    def clear(self):
        """Clears this instance's cache."""
        if self._cache is not None:
            with self._cache as c:
                c.clear()
                c.out_deque.clear()

    def add_response(self, req, resp):
        """Adds the response from sending to `req` to this instance's cache.

        Args:
          req (`ServicecontrolServicesCheckRequest`): the request
          resp (CheckResponse): the response from sending the request
        """
        if self._cache is None:
            return
        signature = sign(req.checkRequest)
        with self._cache as c:
            now = self._timer()
            quota_scale = 0  # WIP
            item = c.get(signature)
            if item is None:
                c[signature] = CachedItem(
                    resp, self.service_name, now, quota_scale)
            else:
                # Update the cached item to reflect that it is updated
                item.last_check_time = now
                item.response = resp
                item.quota_scale = quota_scale
                item.is_flushing = False
                c[signature] = item

    def check(self, req):
        """Determine if ``req`` is in this instances cache.

        Determine if there are cache hits for the request in this aggregator
        instance.

        Not in the cache

        If req is not in the cache, it returns ``None`` to indicate that the
        caller should send the request.

        Cache Hit; response has errors

        When a cached CheckResponse has errors, it's assumed that ``req`` would
        fail as well, so the cached CheckResponse is returned.  However, the
        first CheckRequest after the flush interval has elapsed should be sent
        to the server to refresh the CheckResponse, though until it's received,
        subsequent CheckRequests should fail with the cached CheckResponse.

        Cache behaviour - response passed

        If the cached CheckResponse has no errors, it's assumed that ``req``
        will succeed as well, so the CheckResponse is returned, with the quota
        info updated to the same as requested.  The requested tokens are
        aggregated until flushed.

        Args:
          req (``ServicecontrolServicesCheckRequest``): to be sent to
            the service control service

        Raises:
           ValueError: if the ``req`` service_name is not the same as
             this instances

        Returns:
           ``CheckResponse``: if an applicable response is cached by this
             instance is available for use or None, if there is no applicable
             response

        """
        if self._cache is None:
            return None  # no cache, send request now
        if not isinstance(req, messages.ServicecontrolServicesCheckRequest):
            raise ValueError('Invalid request')
        if req.serviceName != self.service_name:
            logger.error('bad check(): service_name %s does not match ours %s',
                         req.serviceName, self.service_name)
            raise ValueError('Service name mismatch')
        check_request = req.checkRequest
        if check_request is None:
            logger.error('bad check(): no check_request in %s', req)
            raise ValueError('Expected operation not set')
        op = check_request.operation
        if op is None:
            logger.error('bad check(): no operation in %s', req)
            raise ValueError('Expected operation not set')
        if op.importance != messages.Operation.ImportanceValueValuesEnum.LOW:
            return None  # op is important, send request now

        signature = sign(check_request)
        with self._cache as cache:
            logger.debug('checking the cache for %s\n%s', signature, cache)
            item = cache.get(signature)
            if item is None:
                return None  # signal to caller to send req
            else:
                return self._handle_cached_response(req, item)

    def _handle_cached_response(self, req, item):
        with self._cache:  # defensive, this re-entrant lock should be held
            if len(item.response.checkErrors) > 0:
                if self._is_current(item):
                    return item.response

                # There are errors, but now it's ok to send a new request
                item.last_check_time = self._timer()
                return None  # signal caller to send req
            else:
                item.update_request(req, self._kinds)
                if self._is_current(item):
                    return item.response

                if (item.is_flushing):
                    logger.warn('last refresh request did not complete')

                item.is_flushing = True
                item.last_check_time = self._timer()
                return None  # signal caller to send req

    def _is_current(self, item):
        age = self._timer() - item.last_check_time
        return age < self._options.flush_interval


class CachedItem(object):
    """CachedItem holds items cached along with a ``CheckRequest``.

    Thread compatible.

    Attributes:
       response (:class:`messages.CachedResponse`): the cached response
       is_flushing (bool): indicates if it's been detected that item
         is stale, and needs to be flushed
       quota_scale (int): WIP, used to determine quota
       last_check_time (datetime.datetime): the last time this instance
         was checked

    """

    def __init__(self, resp, service_name, last_check_time, quota_scale):
        self.last_check_time = last_check_time
        self.quota_scale = quota_scale
        self.is_flushing = False
        self.response = resp
        self._service_name = service_name
        self._op_aggregator = None

    def update_request(self, req, kinds):
        agg = self._op_aggregator
        if agg is None:
            self._op_aggregator = operation.Aggregator(
                req.checkRequest.operation, kinds)
        else:
            agg.add(req.checkRequest.operation)

    def extract_request(self):
        if self._op_aggregator is None:
            return None

        op = self._op_aggregator.as_operation()
        self._op_aggregator = None
        check_request = messages.CheckRequest(operation=op)
        return messages.ServicecontrolServicesCheckRequest(
            serviceName=self._service_name,
            checkRequest=check_request)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from setuptools import setup, find_packages

# Get the version
version_regex = r'__version__ = ["\']([^"\']*)["\']'
with open('google/api/control/__init__.py', 'r') as f:
    text = f.read()
    match = re.search(version_regex, text)
    if match:
        version = match.group(1)
    else:
        raise RuntimeError("No version number found!")

install_requires = [
    'cachetools>=1.0.0,<2',
    "dogpile.cache>=0.6.1,<0.7",
    'enum34>=1.1.6,<2',
    'google-apitools>=0.4.15,<0.5',
    'oauth2client>=1.5.2,<2',
    'ply>=3.8,<4.0',
    "pylru>=1.0.9,<2.0",
    "pyjwkest>=1.0.0,<=1.0.9",
    "requests>=2.10.0,<3.0",
    'strict-rfc3339>=0.7,<0.8',
    'urllib3>=1.16,<2.0',
]

tests_require = [
    "flask>=0.11.1",
    "httmock>=1.2",
    "mock>=2.0",
    "pytest",
    "pytest-cov"
]

setup(
    name='google-endpoints-api-management',
    version=version,
    description='Google Endpoints API management',
    long_description=open('README.rst').read(),
    author='Google Inc',
    author_email='googleapis-packages@google.com',
    url='https://github.com/cloudendpoints/endpoints-management-python',
    packages=find_packages(),
    namespace_packages=['google'],
    package_dir={'google-endpoints-api-management': 'google'},
    license='Apache License',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    install_requires=install_requires,
    setup_requires=["pytest_runner"],
    tests_require=tests_require,
    test_suite="tests"
)



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import hashlib

import unittest2
from expects import equal, expect

from google.api.control import signing


class TestAddDictToHash(unittest2.TestCase):
    NOTHING_ADDED = hashlib.md5().digest()

    def test_should_add_nothing_when_dict_is_none(self):
        md5 = hashlib.md5()
        signing.add_dict_to_hash(md5, None)
        expect(md5.digest()).to(equal(self.NOTHING_ADDED))

    def test_should_add_matching_hashes_for_matching_dicts(self):
        a_dict = {'test': 'dict'}
        same_dict = dict(a_dict)
        want_hash = hashlib.md5()
        signing.add_dict_to_hash(want_hash, a_dict)
        want = want_hash.digest()
        got_hash = hashlib.md5()
        signing.add_dict_to_hash(got_hash, same_dict)
        got = got_hash.digest()
        expect(got).to(equal(want))



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import datetime
import json
import os
import tempfile
import unittest2

from apitools.base.py import encoding
from expects import be_false, be_none, be_true, expect, equal, raise_error

from google.api.control import messages, service


_LOGGING_DESTINATIONS_INPUT = """
{
  "logs": [{
    "name": "endpoints-log",
    "labels": [{
      "key": "supported/endpoints-log-label"
    }, {
      "key": "unsupported/endpoints-log-label"
    }]
  }, {
    "name": "unreferenced-log",
    "labels": [{
      "key": "supported/unreferenced-log-label"
    }]
  }],

  "monitoredResources": [{
    "type": "endpoints.googleapis.com/endpoints",
    "labels": [{
      "key": "unsupported/endpoints"
    }, {
      "key": "supported/endpoints"
    }]
  }],

  "logging": {
    "producerDestinations": [{
      "monitoredResource": "bad-monitored-resource",
      "logs": [
        "bad-monitored-resource-log"
      ]
    }, {
      "monitoredResource": "endpoints.googleapis.com/endpoints",
      "logs": [
        "bad-endpoints-log",
        "endpoints-log"
      ]
    }]
  }
}

"""


class _JsonServiceBase(object):

    def setUp(self):
        self._subject = encoding.JsonToMessage(messages.Service, self._INPUT)

    def _extract(self):
        return service.extract_report_spec(
            self._subject,
            label_is_supported=fake_is_label_supported,
            metric_is_supported=fake_is_metric_supported
        )

    def _get_registry(self):
        return service.MethodRegistry(self._subject)


class TestLoggingDestinations(_JsonServiceBase, unittest2.TestCase):
    _INPUT = _LOGGING_DESTINATIONS_INPUT
    _WANTED_LABELS = [
        'supported/endpoints-log-label',
        'supported/endpoints'
    ]

    def test_should_access_the_valid_referenced_log(self):
        logs, _metrics, _labels =  self._extract()
        expect(logs).to(equal(set(['endpoints-log'])))

    def test_should_not_specify_any_metrics(self):
        _logs, metrics, _labels =  self._extract()
        expect(metrics).to(equal([]))

    def test_should_specify_the_labels_associated_with_the_valid_log(self):
        _logs, _metrics, labels =  self._extract()
        expect(set(labels)).to(equal(set(self._WANTED_LABELS)))

    def test_should_drop_conflicting_log_labels(self):
        conflicting_label = messages.LabelDescriptor(
            key='supported/endpoints-log-label',
            valueType=messages.LabelDescriptor.ValueTypeValueValuesEnum.BOOL
        )
        bad_log_desc = messages.LogDescriptor(
            name='bad-endpoints-log',
            labels=[conflicting_label]
        )
        self._subject.logs.append(bad_log_desc)
        _logs, _metrics, labels =  self._extract()
        expect(set(labels)).to(equal(set(self._WANTED_LABELS)))



_METRIC_DESTINATIONS_INPUTS = """
{
  "metrics": [{
    "name": "supported/endpoints-metric",
    "labels": [{
      "key": "supported/endpoints-metric-label"
    }, {
      "key": "unsupported/endpoints-metric-label"
    }]
  }, {
    "name": "unsupported/unsupported-endpoints-metric",
    "labels": [{
      "key": "supported/unreferenced-metric-label"
    }]
  }, {
    "name": "supported/non-existent-resource-metric",
    "labels": [{
      "key": "supported/non-existent-resource-metric-label"
    }]
  }],

  "monitoredResources": {
    "type": "endpoints.googleapis.com/endpoints",
    "labels": [{
      "key": "unsupported/endpoints"
    }, {
      "key": "supported/endpoints"
    }]
  },

  "monitoring": {
    "consumerDestinations": [{
      "monitoredResource": "endpoints.googleapis.com/endpoints",
      "metrics": [
        "supported/endpoints-metric",
        "unsupported/unsupported-endpoints-metric",
        "supported/unknown-metric"
      ]
    }, {
      "monitoredResource": "endpoints.googleapis.com/non-existent",
      "metrics": [
         "supported/endpoints-metric",
         "unsupported/unsupported-endpoints-metric",
         "supported/unknown-metric",
         "supported/non-existent-resource-metric"
      ]
    }]
  }
}

"""

class TestMetricDestinations(_JsonServiceBase, unittest2.TestCase):
    _INPUT = _METRIC_DESTINATIONS_INPUTS
    _WANTED_METRICS = [
        'supported/endpoints-metric'
    ]
    _WANTED_LABELS = [
        'supported/endpoints-metric-label',
        'supported/endpoints'
    ]

    def test_should_not_load_any_logs(self):
        logs, _metrics, _labels =  self._extract()
        expect(logs).to(equal(set()))

    def test_should_specify_some_metrics(self):
        _logs, metrics, _labels =  self._extract()
        expect(metrics).to(equal(self._WANTED_METRICS))

    def test_should_specify_the_labels_associated_with_the_metrics(self):
        _logs, _metrics, labels =  self._extract()
        expect(set(labels)).to(equal(set(self._WANTED_LABELS)))


_NOT_SUPPORTED_PREFIX = 'unsupported/'


_COMBINED_LOG_METRIC_LABEL_INPUTS = """
{
  "logs": {
    "name": "endpoints-log",
    "labels": [{
      "key": "supported/endpoints-log-label"
    }, {
      "key": "unsupported/endpoints-log-label"
    }]
  },

  "metrics": [{
    "name": "supported/endpoints-metric",
    "labels": [{
      "key": "supported/endpoints-metric-label"
    }, {
      "key": "unsupported/endpoints-metric-label"
    }]
  }, {
    "name": "supported/endpoints-consumer-metric",
    "labels": [{
      "key": "supported/endpoints-metric-label"
    }, {
      "key": "supported/endpoints-consumer-metric-label"
    }]
  }, {
    "name": "supported/endpoints-producer-metric",
    "labels": [{
      "key": "supported/endpoints-metric-label"
    }, {
      "key": "supported/endpoints-producer-metric-label"
    }]
  }],

  "monitoredResources": {
    "type": "endpoints.googleapis.com/endpoints",
    "labels": [{
      "key": "unsupported/endpoints"
    }, {
      "key": "supported/endpoints"
    }]
  },

  "logging": {
    "producerDestinations": [{
      "monitoredResource": "endpoints.googleapis.com/endpoints",
      "logs": ["endpoints-log"]
    }]
  },

  "monitoring": {
    "consumerDestinations": [{
      "monitoredResource": "endpoints.googleapis.com/endpoints",
      "metrics": [
         "supported/endpoints-consumer-metric",
         "supported/endpoints-metric"
      ]
    }],

    "producerDestinations": [{
      "monitoredResource": "endpoints.googleapis.com/endpoints",
      "metrics": [
         "supported/endpoints-producer-metric",
         "supported/endpoints-metric"
      ]
    }]
  }
}

"""

class TestCombinedExtraction(_JsonServiceBase, unittest2.TestCase):
    _INPUT = _COMBINED_LOG_METRIC_LABEL_INPUTS
    _WANTED_METRICS = [
        "supported/endpoints-metric",
        "supported/endpoints-consumer-metric",
        "supported/endpoints-producer-metric"
    ]
    _WANTED_LABELS = [
        "supported/endpoints",  # from monitored resource
        "supported/endpoints-log-label",  # from log
        "supported/endpoints-metric-label",  # from both metrics
        "supported/endpoints-consumer-metric-label",  # from consumer metric
        "supported/endpoints-producer-metric-label"  # from producer metric
    ]

    def test_should_load_the_specified_logs(self):
        logs, _metrics, _labels =  self._extract()
        expect(logs).to(equal(set(['endpoints-log'])))

    def test_should_load_the_specified_metrics(self):
        _logs, metrics, _labels =  self._extract()
        expect(set(metrics)).to(equal(set(self._WANTED_METRICS)))

    def test_should_load_the_specified_metrics(self):
        _logs, _metrics, labels =  self._extract()
        expect(set(labels)).to(equal(set(self._WANTED_LABELS)))


def fake_is_label_supported(label_desc):
    return not label_desc.key.startswith(_NOT_SUPPORTED_PREFIX)


def fake_is_metric_supported(metric_desc):
    return not metric_desc.name.startswith(_NOT_SUPPORTED_PREFIX)


_NO_NAME_SERVICE_CONFIG_TEST = """
{
    "name": ""
}
"""

class TestBadServiceConfig(_JsonServiceBase, unittest2.TestCase):
    _INPUT = _NO_NAME_SERVICE_CONFIG_TEST

    def test_should_fail_if_service_is_bad(self):
        testfs = [
            lambda: self._get_registry(),
            lambda: service.MethodRegistry(None),
            lambda: service.MethodRegistry(object()),
        ]
        for f in testfs:
            expect(f).to(raise_error(ValueError))


_EMPTY_SERVICE_CONFIG_TEST = """
{
    "name": "empty"
}
"""

class TestEmptyServiceConfig(_JsonServiceBase, unittest2.TestCase):
    _INPUT = _EMPTY_SERVICE_CONFIG_TEST

    def test_should_obtain_a_registry(self):
        expect(self._get_registry()).not_to(be_none)

    def test_lookup_should_return_none(self):
        test_verbs = ('GET', 'POST')
        registry = self._get_registry()
        for v in test_verbs:
            info = registry.lookup('GET', 'any_url')
            expect(info).to(be_none)


_BAD_HTTP_RULE_CONFIG_TEST = """
{
    "name": "bad-http-rule",
    "http": {
        "rules": [{
            "selector": "Uvw.BadRule.DoubleWildCard",
            "get": "/uvw/not_present/**/**"
        }, {
            "selector": "Uvw.BadRule.NoMethod"
        }, {
            "get": "/uvv/bad_rule/no_selector"
        },{
            "selector": "Uvw.OkRule",
            "get": "/uvw/ok_rule/*"
        }]
    }
}
"""

class TestBadHttpRuleServiceConfig(_JsonServiceBase, unittest2.TestCase):
    _INPUT = _BAD_HTTP_RULE_CONFIG_TEST

    def test_lookup_should_return_none_for_unknown_uri(self):
        test_verbs = ('GET', 'POST')
        registry = self._get_registry()
        for v in test_verbs:
            expect(registry.lookup(v, 'uvw/unknown_url')).to(be_none)

    def test_lookup_should_return_none_for_potential_bad_url_match(self):
        registry = self._get_registry()
        expect(registry.lookup('GET', '/uvw/not_present/is_bad')).to(be_none)
        expect(registry.lookup('GET', '/uvw/bad_rule/no_selector')).to(be_none)
        expect(registry.lookup('GET', 'uvw/not_present/is_bad')).to(be_none)
        expect(registry.lookup('GET', 'uvw/not_present/is_bad')).to(be_none)


_USAGE_CONFIG_TEST = """
{
    "name": "usage-config",
    "usage": {
        "rules": [{
            "selector" : "Uvw.Method1",
            "allowUnregisteredCalls" : true
        }, {
            "selector" : "Uvw.Method2",
            "allowUnregisteredCalls" : false
        }, {
            "selector" : "Uvw.IgnoredMethod",
            "allowUnregisteredCalls" : false
        }]
    },
    "http": {
        "rules": [{
            "selector": "Uvw.Method1",
            "get": "/uvw/method1/*"
        }, {
            "selector": "Uvw.Method2",
            "get": "/uvw/method2/*"
        }, {
            "selector": "Uvw.DefaultUsage",
            "get": "/uvw/default_usage"
        }]
    }
}
"""

class TestMethodRegistryUsageConfig(_JsonServiceBase, unittest2.TestCase):
    _INPUT = _USAGE_CONFIG_TEST

    def test_should_detect_with_unregistered_calls_are_allowed(self):
        registry = self._get_registry()
        info = registry.lookup('GET', 'uvw/method1/abc')
        expect(info).not_to(be_none)
        expect(info.selector).to(equal('Uvw.Method1'))
        expect(info.allow_unregistered_calls).to(be_true)

    def test_should_detect_when_unregistered_calls_are_not_allowed(self):
        registry = self._get_registry()
        info = registry.lookup('GET', 'uvw/method2/abc')
        expect(info).not_to(be_none)
        expect(info.selector).to(equal('Uvw.Method2'))
        expect(info.allow_unregistered_calls).to(be_false)

    def test_should_default_to_disallowing_unregistered_calls(self):
        registry = self._get_registry()
        info = registry.lookup('GET', 'uvw/default_usage')
        expect(info).not_to(be_none)
        expect(info.selector).to(equal('Uvw.DefaultUsage'))
        expect(info.allow_unregistered_calls).to(be_false)


_SYSTEM_PARAMETER_CONFIG_TEST = """
{
    "name": "system-parameter-config",
    "systemParameters": {
       "rules": [{
         "selector": "Uvw.Method1",
         "parameters": [{
            "name": "name1",
            "httpHeader": "Header-Key1",
            "urlQueryParameter": "param_key1"
         }, {
            "name": "name2",
            "httpHeader": "Header-Key2",
            "urlQueryParameter": "param_key2"
         }, {
            "name": "api_key",
            "httpHeader": "ApiKeyHeader",
            "urlQueryParameter": "ApiKeyParam"
         }, {
            "httpHeader": "Ignored-NoName-Key3",
            "urlQueryParameter": "Ignored-NoName-key3"
         }]
       }, {
         "selector": "Bad.NotConfigured",
         "parameters": [{
            "name": "neverUsed",
            "httpHeader": "NeverUsed-Key1",
            "urlQueryParameter": "NeverUsed_key1"
         }]
       }]
    },
    "http": {
        "rules": [{
            "selector": "Uvw.Method1",
            "get": "/uvw/method1/*"
        }, {
            "selector": "Uvw.DefaultParameters",
            "get": "/uvw/default_parameters"
        }]
    }
}
"""


class TestMethodRegistrySystemParameterConfig(_JsonServiceBase, unittest2.TestCase):
    _INPUT = _SYSTEM_PARAMETER_CONFIG_TEST

    def test_should_detect_registered_system_parameters(self):
        registry = self._get_registry()
        info = registry.lookup('GET', 'uvw/method1/abc')
        expect(info).not_to(be_none)
        expect(info.selector).to(equal('Uvw.Method1'))
        expect(info.url_query_param('name1')).to(equal(('param_key1',)))
        expect(info.url_query_param('name2')).to(equal(('param_key2',)))
        expect(info.header_param('name1')).to(equal(('Header-Key1',)))
        expect(info.header_param('name2')).to(equal(('Header-Key2',)))

    def test_should_detect_default_api_key(self):
        registry = self._get_registry()
        info = registry.lookup('GET', 'uvw/method1/abc')
        expect(info).not_to(be_none)
        expect(info.api_key_http_header).to(equal(('ApiKeyHeader',)))
        expect(info.api_key_url_query_params).to(equal(('ApiKeyParam',)))

    def test_should_find_nothing_for_unregistered_params(self):
        registry = self._get_registry()
        info = registry.lookup('GET', 'uvw/method1/abc')
        expect(info.url_query_param('name3')).to(equal(tuple()))
        expect(info.header_param('name3')).to(equal(tuple()))

    def test_should_detect_nothing_for_methods_with_no_registration(self):
        registry = self._get_registry()
        info = registry.lookup('GET', 'uvw/default_parameters')
        expect(info).not_to(be_none)
        expect(info.selector).to(equal('Uvw.DefaultParameters'))
        expect(info.url_query_param('name1')).to(equal(tuple()))
        expect(info.url_query_param('name2')).to(equal(tuple()))
        expect(info.header_param('name1')).to(equal(tuple()))
        expect(info.header_param('name2')).to(equal(tuple()))


_BOOKSTORE_CONFIG_TEST = """
{
    "name": "bookstore-http-api",
    "http": {
        "rules": [{
            "selector": "Bookstore.ListShelves",
            "get": "/shelves"
        }, {
            "selector": "Bookstore.CorsShelves",
            "custom": {
               "kind": "OPTIONS",
               "path": "shelves"
            }
        }, {
            "selector": "Bookstore.ListBooks",
            "get": "/shelves/{shelf=*}/books"
        },{
            "selector": "Bookstore.CreateBook",
            "post": "/shelves/{shelf=*}/books",
            "body": "book"
        }]
    }
}
"""

class TestMethodRegistryBookstoreConfig(_JsonServiceBase, unittest2.TestCase):
    _INPUT = _BOOKSTORE_CONFIG_TEST

    def test_configures_list_shelves_ok(self):
        registry = self._get_registry()
        info = registry.lookup('GET', '/shelves')
        expect(info).not_to(be_none)
        expect(info.selector).to(equal('Bookstore.ListShelves'))
        expect(info.body_field_path).to(equal(''))

    def test_configures_options_shelves_ok(self):
        registry = self._get_registry()
        info = registry.lookup('OPTIONS', '/shelves')
        expect(info).not_to(be_none)
        expect(info.selector).to(equal('Bookstore.CorsShelves'))

    def test_configures_list_books_ok(self):
        registry = self._get_registry()
        info = registry.lookup('GET', '/shelves/88/books')
        expect(info).not_to(be_none)
        expect(info.selector).to(equal('Bookstore.ListBooks'))
        expect(info.body_field_path).to(equal(''))

    def test_configures_create_book_ok(self):
        registry = self._get_registry()
        info = registry.lookup('POST', '/shelves/88/books')
        expect(info).not_to(be_none)
        expect(info.selector).to(equal('Bookstore.CreateBook'))
        expect(info.body_field_path).to(equal('book'))


_OPTIONS_SELECTOR_CONFIG_TEST = """
{
    "name": "options-selector",
    "http": {
        "rules": [{
            "selector": "options-selector.OPTIONS",
            "get": "/shelves"
        }, {
            "selector": "options-selector.OPTIONS.1",
            "get": "/shelves/{shelf}"
        }]
    }
}
"""

class TestOptionsSelectorConfig(_JsonServiceBase, unittest2.TestCase):
    _INPUT = _OPTIONS_SELECTOR_CONFIG_TEST

    def test_should_options_to_be_updated(self):
        registry = self._get_registry()
        info = registry.lookup('OPTIONS', '/shelves')
        expect(info).not_to(be_none)
        expect(info.selector).to(equal('options-selector.OPTIONS.2'))


class TestSimpleLoader(unittest2.TestCase):

    def test_should_load_service_ok(self):
        loaded = service.Loaders.SIMPLE.load()
        registry = service.MethodRegistry(loaded)
        info = registry.lookup('GET', '/anything')
        expect(info).not_to(be_none)
        info = registry.lookup('POST', '/anything')
        expect(info).not_to(be_none)


class TestEnvironmentLoader(unittest2.TestCase):

    def setUp(self):
        _config_fd = tempfile.NamedTemporaryFile(delete=False)
        with _config_fd as f:
            f.write(_BOOKSTORE_CONFIG_TEST)
        self._config_file = _config_fd.name
        os.environ[service.CONFIG_VAR] = self._config_file

    def tearDown(self):
        if os.path.exists(self._config_file):
            os.remove(self._config_file)

    def test_does_not_load_if_env_is_not_set(self):
        del os.environ[service.CONFIG_VAR]
        loaded = service.Loaders.ENVIRONMENT.load()
        expect(loaded).to(be_none)

    def test_does_not_load_if_file_is_missing(self):
        os.remove(self._config_file)
        loaded = service.Loaders.ENVIRONMENT.load()
        expect(loaded).to(be_none)

    def test_does_not_load_if_config_is_bad(self):
        with open(self._config_file, 'w') as f:
            f.write('this is not json {')
        loaded = service.Loaders.ENVIRONMENT.load()
        expect(loaded).to(be_none)

    def test_should_load_service_ok(self):
        loaded = service.Loaders.ENVIRONMENT.load()
        expect(loaded).not_to(be_none)
        registry = service.MethodRegistry(loaded)
        info = registry.lookup('GET', '/shelves')
        expect(info).not_to(be_none)


_AUTHENTICATION_CONFIG_TEST = """
{
    "name": "authentication-config",
    "authentication": {
        "rules": [{
            "selector": "Bookstore.ListShelves",
            "requirements": [{
                "providerId": "shelves-provider",
                "audiences": "aud1,aud2"
            }]
        }]
    },
    "http": {
        "rules": [{
            "selector": "Bookstore.ListShelves",
            "get": "/shelves"
        }, {
            "selector": "Bookstore.CorsShelves",
            "custom": {
               "kind": "OPTIONS",
               "path": "shelves"
            }
        }, {
            "selector": "Bookstore.ListBooks",
            "get": "/shelves/{shelf=*}/books"
        },{
            "selector": "Bookstore.CreateBook",
            "post": "/shelves/{shelf=*}/books",
            "body": "book"
        }]
    }
}
"""

class TestAuthenticationConfig(_JsonServiceBase, unittest2.TestCase):
    _INPUT = _AUTHENTICATION_CONFIG_TEST

    def test_lookup_method_with_authentication(self):
      registry = self._get_registry()
      info = registry.lookup('GET', '/shelves')
      auth_info = info.auth_info
      self.assertIsNotNone(auth_info)
      self.assertTrue(auth_info.is_provider_allowed("shelves-provider"))
      self.assertFalse(auth_info.is_provider_allowed("random-provider"))
      self.assertEqual(["aud1", "aud2"],
                       auth_info.get_allowed_audiences("shelves-provider"))
      self.assertEqual([], auth_info.get_allowed_audiences("random-provider"))

    def test_lookup_method_without_authentication(self):
      registry = self._get_registry()
      info = registry.lookup('OPTIONS', '/shelves')
      self.assertIsNotNone(info)
      self.assertIsNone(info.auth_info)
